# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import enum
import time

from tinyolap.cube import Cube


class Slice:
    """Represents a slice from a cube. Slices can be seen as a report in Excel with filters on top,
     row- and column-headers and the requested data itself. See 'samples/web_demo.py' for demo usage."""

    class ColorSchema:

        class Colors(enum.Enum):
            none = ''
            reset = '\033[0m'
            bold = '\033[1m'
            disable = '\033[02m'
            underline = '\033[04m'
            reverse = '\033[07m'
            strikethrough = '\033[09m'
            invisible = '\033[08m'

            black = '\033[30m'
            red = '\033[31m'
            green = '\033[32m'
            orange = '\033[33m'
            blue = '\033[34m'
            purple = '\033[35m'
            cyan = '\033[36m'
            white = '\033[37m'  # aka lightgrey
            darkgrey = '\033[90m'
            lightred = '\033[91m'
            lightgreen = '\033[92m'
            yellow = '\033[93m'
            lightblue = '\033[94m'
            pink = '\033[95m'
            lightcyan = '\033[96m'

            bg_black = '\033[40m'
            bg_red = '\033[41m'
            bg_green = '\033[42m'
            bg_orange = '\033[43m'
            bg_blue = '\033[44m'
            bg_purple = '\033[45m'
            bg_cyan = '\033[46m'
            bg_white = '\033[47m'  # aka lightgrey

        def __init__(self):
            self.colors = self.Colors
            self.borders = ""
            self.negatives = ""
            self.positives = ""
            self.names = ""
            self.members = ""

    class ColorSchemeNone(ColorSchema):
        def __init__(self):
            super().__init__()

    class ColorSchemaDefault(ColorSchema):
        def __init__(self):
            super().__init__()
            self.borders = self.colors.darkgrey
            self.negatives = self.colors.red
            self.positives = self.Colors.none
            self.names = str(self.colors.white) + str(self.colors.bg_red)
            self.members = str(self.colors.blue) + str(self.colors.bold)

    def __init__(self, cube: Cube, definition, suppress_zero_columns=False, suppress_zero_rows=False):
        # self.header_def = {}
        # self.col_def = {}
        # self.rows_def = {}
        self.axis = []
        self.grid = []
        self.grid_rows_count = 0
        self.grid_cols_count = 0
        self.zero_rows = []
        self.zero_cols = []
        self.dimensions = {}
        self.measures = []
        self.cube = cube  # weakref.ref(cube) if cube else None
        self.definition = definition
        self.suppress_zero_columns = suppress_zero_columns
        self.suppress_zero_rows = suppress_zero_rows
        self.title = None

        self.__validate()
        self.__prepare()
        self.refresh()

    def _experimental_refresh_grid(self, parallel_execution=True):
        from concurrent.futures import ThreadPoolExecutor
        from time import time

        results = list()
        addresses = [tuple(arg[5]) + (arg[6],) for arg in self.grid]
        start = time()
        if parallel_execution:
            # note/finding: NOT faster than non-parallel execution
            with ThreadPoolExecutor() as executor:
                for result in executor.map(self._experimental_refresh_grid_cell, addresses):
                    # put results into correct output list:
                    results.append(result)
        else:
            for address in addresses:
                results.append(self._experimental_refresh_grid_cell(address))
        return time() - start, results

    def _experimental_refresh_grid_cell(self, address):
        # arg := [col, row, value, col_members, row_members, idx_address, measure]
        value = self.cube.get(address)
        return value

    def refresh(self):
        """Refreshes the data defined by the slice. Only required when data has changed since the creation
        of the Slice or since the last to 'refresh()'."""
        dim_count = len(self.cube._dimensions)
        address = [""] * dim_count
        measure = ""
        grid = []
        formats = {}

        # set fixed header values
        for member in self.axis[0]:
            dim_ordinal = member[0]
            if dim_ordinal == -1:
                measure = member[1]
            else:
                address[dim_ordinal] = member[1]

        # iterate over columns and rows
        row = 0
        col_members = []
        row_members = []
        for row_member_set in self.axis[2]:
            row_members = []
            for row_member in row_member_set:
                dim_ordinal = row_member[0]
                row_members.append(row_member[1])
                if dim_ordinal == -1:
                    measure = row_member[1]
                else:
                    address[dim_ordinal] = row_member[1]

            col = 0
            for col_member_set in self.axis[1]:
                col_members = []
                for col_member in col_member_set:
                    dim_ordinal = col_member[0]
                    col_members.append(col_member[1])
                    if dim_ordinal == -1:
                        measure = col_member[1]
                    else:
                        address[dim_ordinal] = col_member[1]

                # check formats
                number_format = None
                for idx, member in enumerate(tuple(address)):
                    if member not in formats:
                        member_format = self.cube.get_dimension_by_index(idx).member_get_format(member)
                        if member_format:
                            number_format = member_format
                        formats[member] = member_format
                    else:
                        number_format = formats[member]

                # now we have a valid idx_address to be evaluated
                value = self.cube.get(tuple(address) + (measure,))
                grid.append([col, row, value, col_members, row_members, tuple(address), measure, number_format])

                col += 1
            row += 1

        self.grid = grid
        self.grid_cols_count = len(self.axis[1])
        self.grid_rows_count = len(self.axis[2])

        # execute zero suppression
        self.__zero_suppression()

    def __validate(self):
        """Validates the definition and adds missing information"""

        self.dimensions = {dim.name: False for dim in self.cube._dimensions}

        # 1. add title and description is missing
        if "title" not in self.definition:
            self.definition["title"] = None
        if "description" not in self.definition:
            self.definition["description"] = None

        self.__validate_axis("header")
        self.__validate_axis("columns", True)
        self.__validate_axis("rows", True)

        # add missing dimensions, if required
        if not all([v for v in self.dimensions.values()]):
            if "header" not in self.definition:
                self.definition["header"] = []

            # not all dimensions are defined yet, lets find suitable members
            #  for the missing dimensions and add them to the header.
            for key in self.dimensions:
                if not self.dimensions[key]:  # = if this dimension is not yet defined.
                    dimension = self.cube.get_dimension(key)
                    # use the first top level member, whatever that is and add  it wrapped in a list
                    member_list = [dimension.get_members_by_level(dimension.get_top_level())[0]]
                    self.definition["header"].append({
                        "dimension": key,
                        "member": member_list
                    })

        # add missing measure, if required
        if not self.measures:
            # simply use the first measure and wrap it into a list.
            self.measures = [list(self.cube.get_measures())[0]]
            if "header" not in self.definition:
                self.definition["header"] = []
            self.definition["header"].append({
                "measure": self.measures
            })

        return True

    def __validate_axis(self, axis: str, mandatory: bool = False):
        # 2. process header
        if axis in self.definition:
            if not type(self.definition[axis]) is list:
                raise ValueError(f"Slice axis '{axis}' expected to be of type 'list', "
                                 f"but type is {type(self.definition[axis])}.")
            for position, member_def in enumerate(self.definition[axis]):
                if not type(member_def) is dict:
                    raise ValueError(f"Slice axis '{axis}' expected to contain a list of definitions of type 'dict', "
                                     f"but instead 'dict' type {type(member_def)} was found.")
                if ("dimension" not in member_def) and ("measure" not in member_def):
                    raise ValueError(f"Slice axis '{axis}' expected to contain either keys 'dimension' or 'measure', "
                                     f"but definition '{member_def}' does contain neither.")

                # ******************************************
                # validate dimension definition
                # ******************************************
                if "dimension" in member_def:
                    if "measure" in member_def:
                        raise ValueError(
                            f"Slice axis '{axis}' contains both keys 'dimension' and 'measure' in "
                            f"definition '{member_def}', but only one key is allowed per definition.")
                    dimension = member_def["dimension"]
                    if dimension not in self.dimensions:
                        raise ValueError(
                            f"Slice axis '{axis}' contains a dimension named '{dimension}' in definition "
                            f"'{member_def}' which is not a dimension of cube '{self.cube.name}'.")
                    self.dimensions[dimension] = True

                    # check for member definition
                    if "member" not in member_def:
                        self.definition[axis][position]["member"] = \
                            self.cube.get_dimension(dimension).get_members()  # get all members as list
                    elif (not member_def["member"]) or (member_def["member"] == "*"):
                        self.definition[axis][position]["member"] = \
                            self.cube.get_dimension(dimension).get_members()  # get all members as list
                    elif type(member_def["member"]) is str:
                        # A single member has been defined
                        if not self.cube.get_dimension(dimension).member_exists(member_def["member"]):
                            raise ValueError(
                                f"Slice axis '{axis}' contains unknown member '{member_def['member']}' in "
                                f"definition '{member_def}'"
                                f"'{self.cube.name}'.")
                        self.definition[axis][position]["member"] = [member_def["member"]]  # convert to list
                    elif type(member_def["member"]) is list:
                        # just validate the list
                        for member in member_def["member"]:
                            if not self.cube.get_dimension(dimension).member_exists(member):
                                raise ValueError(f"Slice axis '{axis}' contains unknown member '{member}' in "
                                                 f"definition '{member_def}'.")
                    else:
                        type_name = type(member_def["member"])
                        raise ValueError(f"Slice axis '{axis}' contains an invalid definition '{member_def}'.")

                # ******************************************
                # validate measure definition
                # ******************************************
                elif "measure" in member_def:
                    if "dimension" in member_def:
                        raise ValueError(
                            f"Slice axis '{axis}' contains both keys 'dimension' and 'measure' in "
                            f"definition '{member_def}', but only one key is allowed per definition.")

                    measure = member_def["measure"]
                    if (not measure) or (measure == "*"):
                        self.definition[axis][position]["measure"] = self.cube.measures  # get all measures
                    if type(measure) is str:
                        if measure not in self.cube.measures:
                            raise ValueError(
                                f"Slice axis '{axis}' contains an unknown measure '{measure}' in "
                                f"definition '{member_def}'.")
                        self.definition[axis][position]["measure"] = [measure]
                    if type(measure) is list:
                        for m in measure:
                            if m not in self.cube._measures:
                                raise ValueError(
                                    f"Slice axis '{axis}' contains an unknown measure '{m}' in "
                                    f"definition '{member_def}'.")
                        self.measures = measure
                    else:
                        raise ValueError(f"Slice axis '{axis}' contains an invalid definition '{member_def}'.")

            return self.definition[axis]
        else:
            if mandatory:
                raise ValueError(f"Mandatory slice axis '{axis}' is missing.")

    def __prepare(self):
        """Prepares the slice for execution."""

        # collect all members and assign them to the axes
        axes = ["header", "columns", "rows"]
        self.axis = []
        for axis_index, axis_name in enumerate(axes):
            self.axis.append([])
            for definition in self.definition[axis_name]:
                members = []
                if "dimension" in definition:
                    dimension = definition["dimension"]
                    dim_ordinal = self.cube.get_dimension_ordinal(dimension)
                    for member in definition["member"]:
                        members.append((dim_ordinal, member, dimension))
                elif "measure" in definition:
                    for measure in definition["measure"]:
                        members.append((-1, measure))
                    pass
                else:
                    raise ValueError("Unexpect error in Slice preparation.")
                self.axis[axis_index].append(members)

            # expand (multiply) all definition of the axis
            if axis_index == 0:
                # only valid for header
                # print(f"{axis_name}:")
                members = [x[0] for x in self.axis[axis_index]]
                # print(members)
                self.axis[axis_index] = members
            else:
                # only valid for rows and columns
                no_of_definitions = len(self.definition[axis_name])
                members = []

                if no_of_definitions == 1:
                    # print(f"{axis_name}:")
                    for t in [[m0]
                              for m0 in self.axis[axis_index][0]]:
                        # print(t)
                        members.append(t)
                elif no_of_definitions == 2:
                    # print(f"{axis_name}:")
                    for t in [[m0, m1]
                              for m0 in self.axis[axis_index][0]
                              for m1 in self.axis[axis_index][1]]:
                        # print(t)
                        members.append(t)
                elif no_of_definitions == 3:
                    # print(f"{axis_name}:")
                    for t in [[m0, m1, m2]
                              for m0 in self.axis[axis_index][0]
                              for m1 in self.axis[axis_index][1]
                              for m2 in self.axis[axis_index][2]]:
                        # print(t)
                        members.append(t)
                elif no_of_definitions == 4:
                    # print(f"{axis_name}:")
                    for t in [[m0, m1, m2, m3]
                              for m0 in self.axis[axis_index][0]
                              for m1 in self.axis[axis_index][1]
                              for m2 in self.axis[axis_index][2]
                              for m3 in self.axis[axis_index][3]]:
                        # print(t)
                        members.append(t)
                elif no_of_definitions == 5:
                    # print(f"{axis_name}:")
                    for t in [[m0, m1, m2, m3, m4]
                              for m0 in self.axis[axis_index][0]
                              for m1 in self.axis[axis_index][1]
                              for m2 in self.axis[axis_index][2]
                              for m3 in self.axis[axis_index][3]
                              for m4 in self.axis[axis_index][4]]:
                        # print(t)
                        members.append(t)
                else:
                    raise ValueError(
                        f"Too many nested dimensions on axis '{axis_name}'. Max 5 nested dimension are allowed.")

                self.axis[axis_index] = members

    def __zero_suppression(self):
        # todo: implementation missing
        self.zero_cols = None
        self.zero_rows = None

    def as_console_output(self, color_schema: ColorSchema = ColorSchema, hide_zeros: bool = True) -> str:
        """Renders the slice suitable for console output. The output contains
        control characters and color definitions and is therefore not suitable
        for other use cases."""

        cell_width = 14
        row_header_width = 16

        row_dims = len(self.grid[0][4])
        col_dims = len(self.grid[0][3])

        text = "\n"

        # title
        if self.definition["title"] or self.title:
            title = self.definition["title"] if not self.title else self.title
            text += ("-" * 80) + "\n"
            text += f"{title}\n"
            if self.definition["description"]:
                text += f"{self.definition['description']}\n"
            text += ("-" * 80) + "\n"

        # header dimensions
        for member in self.axis[0]:
            if member[0] == -1:
                text += f"Measure := {member[1]}\n"
            else:
                text += f"{member[2]} := {member[1]}\n"

        # col headers
        for c in range(col_dims):
            for r in range(row_dims):
                text += " ".ljust(row_header_width)
            for i in range(self.grid_cols_count):
                caption = self.grid[i][3][c]
                if len(caption) > cell_width:
                    caption = caption[:cell_width - 3].strip() + "..."
                text += caption.center(cell_width)
            text += "\n"

        # row headers & cells
        previous = {}
        for cell in self.grid:
            col = cell[0]
            row = cell[1]
            value = cell[2]
            format = cell[7]
            if type(value) is float:
                if hide_zeros and value == 0.0:
                    value = f"-".rjust(cell_width)
                else:
                    if format:
                        value = format.format(value).rjust(cell_width)
                    else:
                        value = f"{value:,.2f}".rjust(cell_width)
            elif value is None:
                value = f"-".rjust(cell_width)
            else:
                value = f"{str(value)}".rjust(cell_width)

            if col == 0:
                if row > 0:
                    text += "\n"
                for pos, member in enumerate(cell[4]):
                    caption = member
                    if len(caption) > row_header_width:
                        caption = caption[:row_header_width - 3].strip + "..."
                    if pos in previous:
                        if previous[pos] != member:
                            text += caption.ljust(row_header_width)
                        else:
                            text += " ".ljust(row_header_width)
                    else:
                        text += caption.ljust(row_header_width)
                    previous[pos] = member

            text += value

        return text

    def as_html(self, footer="", dropdowns=None) -> str:
        """Renders the slice as a complete html page. Not suitable for other purposes.
        This is a hardcoded quick and dirty implementation for use in the 'samples/web_demo.py'."""
        start = time.time()

        row_dims = len(self.grid[0][4])
        col_dims = len(self.grid[0][3])

        tro = "<tr>"
        trc = "</tr>\n"

        text = ""
        # title
        if self.definition["title"] or self.title:
            title = self.definition["title"] if not self.title else self.title
            text += f"<h2>{title}</h2>\n"
            if self.definition["description"]:
                text += f"<h4>{self.definition['description']}</h4>\n"
        text += f'<div class="font-italic font-weight-light">{footer}</div>'

        # header dimensions
        text += '<table class="table w-auto"><tbody>\n'
        for member in self.axis[0]:
            if member[0] == -1:
                text += f'<tr><th scope="row">Measure</th><td>{member[1]}</td></tr>\n'
            else:
                text += f'<tr><th scope="row">{member[2]}</th><td>{member[1]}</td></tr>\n'
        text += '</tbody></table>'

        text += '<div style= width: 100%">'
        text += '<div class"table-responsive">' \
                '<table class="table table-hover table-striped table-bordered"' \
                '>\n'
        text += '<thead">\n'

        # column headers
        text += tro
        for c in range(col_dims):
            for r in range(row_dims):
                text += f'<th scope="col" class="th-lg" style="width: 80px"></th>\n'
                # text += tdo + " ".ljust(row_header_width) + tdc
            for i in range(self.grid_cols_count):
                text += f'<th scope="col" class="text-center" style="width: 80px">' \
                        f'{self.grid[i][3][c]}' \
                        f'</th>\n'
        text += trc
        text += '</thead">\n'

        # row headers and cells
        previous = {}
        text += tro
        for cell in self.grid:
            col = cell[0]
            row = cell[1]
            value = cell[2]
            format = cell[7]

            # row headers
            if col == 0:
                if row > 0:
                    text += trc
                    text += tro
                for pos, member in enumerate(cell[4]):
                    if pos in previous:
                        if previous[pos] != member:
                            text += f'<th class="text-nowrap" scope="row">{member}</th>\n'
                        else:
                            text += f'<th class="text-nowrap" scope="row"></th>\n'
                    else:
                        text += f'<th class="text-nowrap" scope="row">{member}</th>\n'
                    previous[pos] = member

            if type(value) is float:
                if format:
                    value = format.format(value)
                else:
                    value = f"{value:,.0f}"  # :,.2f}"
            elif value is None:
                value = "-"
            else:
                value = ""
            text += f'<td class="text-nowrap" style="text-align: right">{value}</td>\n'

        text += trc
        text += "</table></div>"
        text += '</div>'

        style = 'table.table-fit {width: auto !important;table-layout: auto !important;}' \
                'table.table-fit thead th, table.table-fit tfoot th {width: auto !important;}' \
                'table.table-fit tbody td, table.table-fit tfoot td {width: auto !important;}'

        header = '<header class="p-3 bg-dark text-white">' \
                 '<div class="container">' \
                 '<div class="d-flex flex-wrap align-items-center justify-content-center justify-content-lg-start">' \
                 '<a href="/" class="d-flex align-items-center mb-2 mb-lg-0 text-white text-decoration-none">' \
                 '<img width="48" height="48" src="/logo.png" aria-label="TinyOlap"></img>' \
                 ' <a class="navbar-brand px-2 text-white">TinyOlap</a>' \
                 '</a>' \
                 '<ul class="nav col-12 col-lg-auto me-lg-auto mb-2 justify-content-center mb-md-0">' \
                 '<li><a href="#" class="nav-link px-2 text-secondary">Home</a></li>' \
                 '<li><a href="#" class="nav-link px-2 text-white">Foo</a></li>' \
                 '<li><a href="#" class="nav-link px-2 text-white">Bar</a></li>' \
                 '<li><a href="#" class="nav-link px-2 text-white">Baz</a></li>' \
                 '</ul>' \
                 '<form action="/report" class="col-12 col-lg-auto mb-3 mb-lg-0 me-lg-3">' \
                 '<div class="btn-toolbar">'\
                 '<button type="submit" class="btn btn-warning">Refresh</button>' \
                 '</div>' \
                 '</form>' \
                 '<form action="/nextreport" class="col-12 col-lg-auto mb-3 mb-lg-0 me-lg-3">' \
                 '<div class="btn-toolbar">'\
                 '<button type="submit" class="btn btn-warning">Next</button>' \
                 '</div>' \
                 '</form>' \
                 '</div>' \
                 '</div>' \
                 '</header>'

        # '<button onClick="Refresh(10)" class="btn btn-warning">Refresh 10x</button>'
        script = 'function Refresh(refreshesLeft) {' \
                 'refreshesLeft = refreshesLeft || 5;' \
                 'window.parent.location = window.parent.location.href + \'?refreshesLeft = \'+refreshesLeft;' \
                 '}' \
                 'function onLoad() {' \
                 'let params = new URLSearchParams(document.location.search.substring(1));' \
                 'let rl = params.get("refreshesLeft")' \
                 'if (rl) Refresh( (rl | 0 ) -1)' \
                 '}' \
                 'document.addEventListener("DOMContentLoaded", onLoad)'

        duration = f'<div class="font-italic font-weight-light">HTML rendered in {time.time()- start:.6} sec.</div>'

        html = f'<!doctype html><html lang="en"><head><!-- Required meta tags --><meta charset="utf-8">' \
               f'<meta name="viewport" content="width=device-width, initial-scale=1">' \
               f'<link href="https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/css/bootstrap.min.css" rel="stylesheet" ' \
               f'integrity="sha384-1BmE4kWBq78iYhFldvKuhfTAU6auU8tT94WrHftjDbrCEXSU1oBoqyl2QvZ6jIW3" ' \
               f'crossorigin="anonymous">' \
               f'<title>TinyOlap report</title>' \
               f'<style>{style}</style>' \
               f'<script>{script}</script>' \
               f'</head>' \
               f'<body>' \
               f'{header}' \
               f'<div class="p-3">' \
               f'{text}' \
               f'</div>' \
               f'{duration}' \
               f'<script src="https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/js/bootstrap.bundle.min.js" ' \
               f'integrity="sha384-ka7Sk0Gln4gmtz2MlQnikT1wXgYsOg+OMhuP+IlRH9sENBO0LRn5q+8nbTov4+1p" ' \
               f'crossorigin="anonymous"></script></body></html>'

        return html

    def as_csv(self) -> str:
        return str(self)

    def __str__(self):
        return self.as_console_output(color_schema=Slice.ColorSchemaDefault())

    def __repr__(self):
        return self.as_console_output(color_schema=Slice.ColorSchemaDefault())



# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import datetime
from abc import ABC, abstractmethod
from collections import Iterable


class Command(ABC):
    """Command base class for TinyOlap commands."""
    def __init__(self):
        super().__init__()
        self.timestamp = datetime.datetime.now()
        self.undone: bool = False

    @abstractmethod
    def undo(self, database) -> bool:
        """Undo the command."""
    @abstractmethod
    def redo(self, database) -> bool:
        """Redo the command."""


class CompoundCommand(Command):
    """Command that represents a list of related commands,
     e.g. a multi-value transaction."""
    def __init__(self, commands = None):
        super().__init__()
        self.undone = False
        self.commands: list[Command] = []
        if isinstance(commands, Iterable):
            self.commands += commands
        else:
            self.commands.append(commands)

    def undo(self, database) -> bool:
        """Undo the command."""
        result = True
        for command in self.commands:
            result &= command.undo(database)
        self.undone = True
        return result

    def redo(self, database) -> bool:
        """Redo the command."""
        result = True
        for command in self.commands:
            result &= command.redo(database)
        self.undone = False
        return result

    def add(self, command:Command):
        self.commands.append(command)

    def __len__(self):
        return len(self.commands)

    

class CubeSetCommand(Command):
    """Command to set a cube value."""
    def __init__(self, cube: str, idx_address: tuple[int], before, after):
        super().__init__()
        self.undone = False
        self.cube = cube
        self.idx_address = idx_address
        self.before = before
        self.after = after

    def undo(self, database) -> bool:
        """Undo the command."""
        database.cubes[self.cube]._set_base_level_cell(self.idx_address, 0, self.before)
        self.undone = True
        return True

    def redo(self, database) -> bool:
        """Redo the command."""
        database.cubes[self.cube]._set_base_level_cell(self.idx_address, 0, self.after)
        self.undone = False
        return True


class AttributeSetCommand(Command):
    """Command to set a dimension attribute."""
    def __init__(self, dimension: str, attribute, member: str, before, after):
        super().__init__()
        self.undone = False
        self.dimension = dimension
        self.attribute = attribute
        self.member = member
        self.before = before
        self.after = after

    def undo(self, database) -> bool:
        """Undo the command."""
        database.dimensions[self.dimension].set_attribute(self.attribute, self.member, self.before)
        self.undone = True
        return True

    def redo(self, database) -> bool:
        """Redo the command."""
        database.dimensions[self.dimension].set_attribute(self.attribute, self.member, self.after)
        self.undone = False
        return True




# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

class Role:
    pass


class Viewer(Role):
    """
    Permissions for read-only actions that do not affect state of a database,
    such as reading (but not modifying) data from cubes.
    """
    pass


class Editor(Role):
    """
    All viewer permissions, plus permissions to modify data in cubes, such as writing
    values to cubes, splashing of values over cubes, importing data and using the history.
    """
    pass


class Owner(Role):
    """
    All editor permissions, plus permissions to modify the structure of the database,
    such adding/modifying/removing of dimension, subsets, attributes, cubes and rules.
    """
    pass


class Admin(Role):
    """All owner permissions, plus permission to add/modify/remove users and roles."""
    pass


class User:
    def __init__(self, name: str, password: str = None, role: Role = Admin()):
        self._name = name
        self._password = password
        self._role = role

    @property
    def name(self) -> str:
        return self._name

    def rename(self, new_name: str):
        self._name = new_name


    @property
    def role(self) -> Role:
        return self._role

    @role.setter
    def role(self, value: Role):
        self._role = value

    @property
    def password(self) -> str:
        return self._password

    @password.setter
    def password(self, value: str):
        self._password = value



# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations
import functools
import inspect

from tinyolap.rules import RuleScope, RuleInjectionStrategy


def rule(cube: str, trigger: list[str], scope: RuleScope = RuleScope.ALL_LEVELS,
         injection: RuleInjectionStrategy = RuleInjectionStrategy.NO_INJECTION,
         volatile: bool = False, command=None):
    """
    Decorator for TinyOlap rule functions.

    :param cube: The cube the rule should be assigned to.
    :param trigger: The cell trigger that should trigger the rule. Either a single member name or a list
                    of member names from different dimensions.
    :param scope: The scope of the rule. Please refer the documentation for further details.
    :param injection: THe injection strategy for the rule
    :param volatile: (optional, default = False) Identifies that the rule may or will return changing
                     results on identical input, e.g. if real-time data integration is used.
    :param command: (optional, default = None) Identifies that this rule will only the triggered
                     and executed by an explicit user command or call to the ``execute()``method of
                     a Cell object.
    """

    def decorator_rule(func):
        @functools.wraps(func)
        def wrapper_rule(*args, **kwargs):
            #args = str(inspect.signature(func))
            #args = args[1: len(args) - 1].split(",")
            return func(*args, **kwargs)

        wrapper_rule.cube = cube
        wrapper_rule.trigger = trigger
        wrapper_rule.scope = scope
        wrapper_rule.injection = injection
        wrapper_rule.volatile = volatile
        wrapper_rule.command = command
        return wrapper_rule

    return decorator_rule




# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

from collections import Iterable
from typing import SupportsFloat

from tinyolap.exceptions import *
from tinyolap.member import Member


# noinspection PyProtectedMember


class Cell(SupportsFloat):
    """
    A Cell is an immutable pointer to a data cell in a cube. Cell objects can
    be used to navigate through data space. In addition they can be directly
    used in mathematical calculations, as they (almost) behave like a 'float' value.

    .. note::
        Cell objects are also used for the internal rules engine of TinyOlap. They are perfect for being
        handed in to function and methods, as shown in the code fragment below.

        .. code:: python

            from tinyolap.database import Database
            from tinyolap.cell import Cell

            # setup a new database
            cell = cube.create_cell()
            value = cell.value
            idx_address = cell.idx_address  # returns a list e.g. ["member of dim1", "member of dim2" ...]
            cell.move("dim1", move.NEXT)  # move.NEXT
    """

    #: Indicates that either subsequent rules should continue and do the calculation
    #: work or that the cell value, either from a base-level or an aggregated cell,
    #: form the underlying cube should be used.
    CONTINUE = object()
    #: Indicates that rules function was not able return a proper result (why ever).
    NONE = object()
    #: Indicates that the rules functions run into an error. Such errors will be
    #: pushed up to initially calling cell request.
    ERROR = object()

    #: Indicates that the rules functions wants to bypass all rules and get the
    #: underlying base_level or aggregated value from the cube.
    BYPASS_RULES = object()

    # region Initialization
    @classmethod
    def create(cls, cube, dim_names, address, bolt):
        cell = Cell()
        cell._cube = cube
        cell._dim_names = dim_names
        cell._dim_count = len(dim_names)
        cell._address = address
        cell._bolt = bolt
        return cell

    def __init__(self):
        self._cube = None
        self._dim_names = None
        self._address = None
        self._bolt = None
        self._dim_count = -1
        pass

    def __new__(cls):
        return SupportsFloat.__new__(cls)

    # endregion

    # region Properties
    @property
    def value(self):
        """Reads the value of the current cell idx_address from the underlying cube."""
        return self._cube._get(self._bolt)

    @value.setter
    def value(self, value):
        """Writes a value of the current cell idx_address to the underlying cube."""
        self._cube._set(self._bolt, value)

    @property
    def numeric_value(self) -> float:
        """Reads the numeric value of the current cell idx_address from the underlying cube."""
        value = self._cube._get(self._bolt)
        if isinstance(value, (int, float, complex)) and not isinstance(value, bool):
            return float(value)
        else:
            return 0.0

    # endregion

    # region methods
    def alter(self, *args) -> Cell:
        """
        Creates a modified deep copy of a Cell object.

        :param args: One or more modifiers (member names) for at least one dimension of the cube.
            Member names can be in one of the ƒfollowing formats:

               {member} e.g.: clone = c.alter("Mar", ...)
               {cube_name:member} e.g.: clone = c.alter("months:Mar", ...)
               {dimension_index:member} e.g.: clone = c.alter("1:Mar", ...)

            If multiple modifiers for a single dimension are defined, then the last of those will be used.
        :return: A new Cell object.
        """
        modifiers = []

        key_level = 6  # LEVEL
        key_name = 1  # NAME
        super_level, idx_address, idx_measure = self._bolt
        idx_address = list(idx_address)
        address = list(self._address)

        for member in args:
            if type(member) is Member:
                # The easy way! The Member object should be properly initialized already.
                raise NotImplementedError("Working on that...")

            elif type(member) is str:
                idx_dim, idx_member, member_level = self._get_member(member)
                address[idx_dim] = self._cube._dimensions[idx_dim].members[idx_member][key_name]

                # adjust the super_level
                super_level -= self._cube._dimensions[idx_dim].members[self._bolt[1][idx_dim]][key_level]
                super_level += member_level

                modifiers.append((idx_dim, idx_member))
            else:
                raise TypeError(f"Invalid type '{type(member)}'. Only type 'str' and 'Member are supported.")

        for modifier in modifiers:
            idx_address[modifier[0]] = modifier[1]
        bolt = (super_level, tuple(idx_address), idx_measure)
        return Cell.create(self._cube, self._dim_names, address, bolt)

    def member(self, dimension_and_or_member_name: str) -> Member:
        """
        Create a new Member object from the Cell's current context.
        This can be either the current member of a specific dimension,
        if called with a dimension name, e.g., ``c.member("months")`` .
        Or it can be already a modified member of a dimension, if called
        with a member name, e.g., ``c.member("months:Jul")`` .

        :param dimension_and_or_member_name: Name of the dimension and or member to be returned.

        .. code:: python

            cell = cube.cell("2022", "Jan", "North", "Sales")
            jan = c.member("months")
            jul = c.member("months:Jul")
            jul = c.member("1:Aug")  # 'months' is the second dimension of the cube, so zero-based index is 1.

        :return: A new Member object.
        """
        idx_dim, idx_member, member_level = self._get_member(dimension_and_or_member_name)
        member = Member(self._cube._dimensions[idx_dim], dimension_and_or_member_name, self._cube, idx_dim, idx_member,
                        member_level)
        return member

    # endregion

    # region Cell manipulation via indexing/slicing
    def __getitem__(self, args):
        if not isinstance(args, (str, Member)) and (args[-1] == self.BYPASS_RULES):
            return self._cube._get(self.__item_to_bold(args[:len(args) - 1]), True)
        else:
            return self._cube._get(self.__item_to_bold(args))

    def __setitem__(self, args, value):
        self._cube._set(self.__item_to_bold(args), value)

    def __delitem__(self, args):
        self.__setitem__(args, None)

    # endregion

    # region Cell manipulation
    def __item_to_bold(self, item):
        """Setting a value through indexing/slicing will temporarily modify the cell idx_address and return
        the value from that cell idx_address. This does NOT modify the cell idx_address of the Cell object.
        To modify the cell idx_address of a Cell, you can call the ``.alter(...)`` method."""

        modifiers = []
        if type(item) is str or not isinstance(item, Iterable):
            item = (item,)

        key_level = 6  # LEVEL
        super_level, idx_address, idx_measure = self._bolt
        idx_address = list(idx_address)

        for member in item:
            if type(member) is Member:
                super_level -= self._cube._dimensions[member._idx_dim].members[self._bolt[1][member._idx_dim]][
                    key_level]
                super_level += member._member_level
                modifiers.append((member._idx_dim, member._idx_member))

            elif type(member) is str:
                idx_dim, idx_member, member_level = self._get_member(member)
                super_level -= self._cube._dimensions[idx_dim].members[self._bolt[1][idx_dim]][key_level]
                super_level += member_level

                modifiers.append((idx_dim, idx_member))
            else:
                raise TypeError(f"Invalid type '{type(member)}'. Only type 'str' and 'Member are supported.")

        # Finally modify the idx_address and set the value
        for modifier in modifiers:
            idx_address[modifier[0]] = modifier[1]
        bolt = (super_level, tuple(idx_address), idx_measure)
        return bolt

    def _get_member(self, member):
        # The hard way! We need to evaluate where the member is coming from
        # Convention: member names come in one of the following formats:
        #   c["Mar"] = 333.0
        #   c["months:Mar"] = 333.0
        #   c["1:Mar"] = 333.0

        level = self._cube._dimensions[0].LEVEL
        dimensions = self._cube._dimensions
        idx_dim = -1
        pos = member.find(":")
        if pos != -1:
            # lets extract the dimension name and check if it is valid, e.g., c["months:Mar"]
            dim_name = member[:pos].strip()

            # special test for ordinal dim position instead of dim name, e.g., c["1:Mar"] = 333.0
            if dim_name.isdigit():
                ordinal = int(dim_name)
                if 0 <= ordinal < len(self._dim_names):
                    # that's a valid dimension position number
                    idx_dim = ordinal
            if idx_dim == -1:
                if dim_name not in self._cube._dim_lookup:
                    raise InvalidCellAddressException(f"Invalid member key. '{dim_name}' is not a dimension "
                                                  f"in cube '{self._cube.name}. Found in '{member}'.")
                idx_dim = self._cube._dim_lookup[dim_name]

            # adjust the member name
            member = member[pos + 1:].strip()
            if member not in dimensions[idx_dim]._member_idx_lookup:
                raise InvalidCellAddressException(f"Invalid member key. '{member}'is not a member of "
                                              f"dimension '{dim_name}' in cube '{self._cube.name}.")
            idx_member = dimensions[idx_dim]._member_idx_lookup[member]

            member_level = dimensions[idx_dim].members[idx_member][self._cube._dimensions[0].LEVEL]
            return idx_dim, idx_member, member_level

        # No dimension identifier in member name, search all dimensions
        # ...we'll search in reverse order, as we assume that it is more likely,
        #    that inner dimensions are requested through rules.
        for idx_dim in range(self._dim_count - 1, -1, -1):
            if member in dimensions[idx_dim]._member_idx_lookup:
                idx_member = dimensions[idx_dim]._member_idx_lookup[member]
                # adjust the super_level
                member_level = dimensions[idx_dim].members[idx_member][level]
                return idx_dim, idx_member, member_level

        # Still nothing found ? Then it might be just a dimension name or dimension ordinal
        # to reference the current member of that dimension.
        dim_name = member
        # special test for ordinal dim position instead of dim name, e.g., c["1:Mar"] = 333.0
        if type(dim_name) is int or dim_name.isdigit():
            ordinal = int(dim_name)
            if 0 <= ordinal < len(self._dim_names):
                # that's a valid dimension position number
                idx_dim = ordinal
            if idx_dim == -1:
                if dim_name not in self._cube._dim_lookup:
                    raise InvalidCellAddressException(f"Invalid member key. '{dim_name}' is not a dimension "
                                                  f"in cube '{self._cube.name}. Found in '{member}'.")
                idx_dim = self._cube._dim_lookup[dim_name]

            idx_member = self._bolt[1][idx_dim]
            member_level = dimensions[idx_dim].members[idx_member][level]
            return idx_dim, idx_member, member_level

        # You loose...
        if idx_dim == -1:
            raise KeyError(f"'{member}' is not a member of any dimension in "
                           f"cube '{self._cube.name}', or a valid reference to any of it's dimensions.")

    # endregion

    # region operator overloading and float behaviour
    def __float__(self) -> float:  # type conversion to float
        return self.numeric_value

    def __index__(self) -> int:  # type conversion to int
        return int(self.numeric_value)

    def __neg__(self):  # - unary operator
        return - self.numeric_value

    def __pos__(self):  # + unary operator
        return self.numeric_value

    def __add__(self, other):  # + operator
        return self.numeric_value + other

    def __iadd__(self, other):  # += operator
        return self.numeric_value + other.numeric_value

    def __radd__(self, other):  # + operator
        return other + self.numeric_value

    def __sub__(self, other):  # - operator
        return self.numeric_value - other

    def __isub__(self, other):  # -= operator
        return self.numeric_value - other

    def __rsub__(self, other):  # - operator
        return other - self.numeric_value

    def __mul__(self, other):  # * operator
        return self.numeric_value * other

    def __imul__(self, other):  # *= operator
        return self.numeric_value * other

    def __rmul__(self, other):  # * operator
        return other * self.numeric_value

    def __floordiv__(self, other):  # // operator (returns an integer)
        return self.numeric_value // other

    def __ifloordiv__(self, other):  # //= operator (returns an integer)
        return self.numeric_value // other

    def __rfloordiv__(self, other):  # // operator (returns an integer)
        return other // self.numeric_value

    def __truediv__(self, other):  # / operator (returns an float)
        return self.numeric_value / other

    def __idiv__(self, other):  # /= operator (returns an float)
        return self.numeric_value / other

    def __rtruediv__(self, other):  # / operator (returns an float)
        return other / self.numeric_value

    def __mod__(self, other):  # % operator (returns a tuple)
        return self.numeric_value % other

    def __imod__(self, other):  # %= operator (returns a tuple)
        return self.numeric_value % other

    def __rmod__(self, other):  # % operator (returns a tuple)
        return other % self.numeric_value

    def __divmod__(self, other):  # div operator (returns a tuple)
        return divmod(self.numeric_value, other)

    def __rdivmod__(self, other):  # div operator (returns a tuple)
        return divmod(other, self.numeric_value)

    def __pow__(self, other, modulo=None):  # ** operator
        return self.numeric_value ** other

    def __ipow__(self, other, modulo=None):  # **= operator
        return self.numeric_value ** other

    def __rpow__(self, other, modulo=None):  # ** operator
        return other ** self.numeric_value

    def __lt__(self, other):  # < (less than) operator
        return self.numeric_value < other

    def __gt__(self, other):  # > (greater than) operator
        return self.numeric_value > other

    def __le__(self, other):  # <= (less than or equal to) operator
        return self.numeric_value <= other

    def __ge__(self, other):  # >= (greater than or equal to) operator
        return self.numeric_value >= other

    def __eq__(self, other):  # == (equal to) operator
        return self.numeric_value == other

    def __and__(self, other):  # and (equal to) operator
        return self.numeric_value and other

    def __iand__(self, other):  # and (equal to) operator
        return self.numeric_value and other

    def __rand__(self, other):  # and (equal to) operator
        return other and self.numeric_value

    def __or__(self, other):  # or (equal to) operator
        return self.numeric_value or other

    def __ior__(self, other):  # or (equal to) operator
        return self.numeric_value or other

    def __ror__(self, other):  # or (equal to) operator
        return other or self.numeric_value

    # endregion

    # region conversion function
    def __abs__(self):
        return self.numeric_value.__abs__()

    def __bool__(self):
        return self.numeric_value.__bool__()

    def __str__(self):
        return self.value.__str__()

    def __int__(self):
        return self.numeric_value.__int__()

    def __ceil__(self):
        return self.numeric_value.__ceil__()
    # endregion



# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations
from itertools import chain

_RaiseKeyError = object() # singleton for no-default behavior

class CaseInsensitiveDict(dict):
    # dicts take a mapping or iterable as their optional first argument
    __slots__ = ()  # no __dict__ - that would be redundant

    @staticmethod  # because this doesn't make sense as a global function.
    def _process_args(mapping=(), **kwargs):
        if hasattr(mapping, 'items'):
            mapping = getattr(mapping, 'items')()
        return ((k.lower(), v) for k, v in chain(mapping, getattr(kwargs, 'items')()))

    def __init__(self, mapping=(), **kwargs):
        super(CaseInsensitiveDict, self).__init__(self._process_args(mapping, **kwargs))

    def __getitem__(self, k):
        return super(CaseInsensitiveDict, self).__getitem__(k.lower())
    def lookup(self, k):
        return super(CaseInsensitiveDict, self).get(k.lower())
    def lookuptry(self, k):
        try:
            return super(CaseInsensitiveDict, self).__getitem__(k.lower())
        except:
            return None

    def __setitem__(self, k, v):
        return super(CaseInsensitiveDict, self).__setitem__(k.lower(), v)

    def __delitem__(self, k):
        return super(CaseInsensitiveDict, self).__delitem__(k.lower())

    def get(self, k, default=None):
        return super(CaseInsensitiveDict, self).get(k.lower(), default)

    def setdefault(self, k, default=None):
        return super(CaseInsensitiveDict, self).setdefault(k.lower(), default)

    def pop(self, k, v=_RaiseKeyError):
        if v is _RaiseKeyError:
            return super(CaseInsensitiveDict, self).pop(k.lower())
        return super(CaseInsensitiveDict, self).pop(k.lower(), v)

    def update(self, mapping=(), **kwargs):
        super(CaseInsensitiveDict, self).update(self._process_args(mapping, **kwargs))

    def __contains__(self, k):
        return super(CaseInsensitiveDict, self).__contains__(k.lower())

    def copy(self):  # don't delegate w/ super - dict.copy() -> dict :(
        return type(self)(self)

    def populate(self, dictionary: dict) -> CaseInsensitiveDict:
        for k, v in dictionary.items():
            self[k] = v
        return self

    @classmethod
    def fromkeys(cls, keys, v=None):
        return super(CaseInsensitiveDict, cls).fromkeys((k.lower() for k in keys), v)

    def __repr__(self):
        return '{0}({1})'.format(type(self).__name__, super(CaseInsensitiveDict, self).__repr__())


# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations
import logging
import os
import sqlite3
from collections.abc import Iterable
from os import path
from pathlib import Path
from timeit import default_timer as timer

from tinyolap.exceptions import *


# noinspection SqlNoDataSourceInspection
class Backend:
    """SQLite storage storage_provider"""
    META_TABLE_CUB = "meta_cub"
    META_TABLE_DIM = "meta_dim"
    META_TABLE_FIELDS = [("key", "TEXT PRIMARY KEY"), ("config", "TEXT")]
    CUB_PREFIX = "cub_"
    DB_FOLDER_NAME = "db"
    DB_EXTENSION = ".db"

    LOG_EXTENSION = ".log"
    LOG_LEVEL = logging.INFO

    def __init__(self, database_name: str, in_memory: bool = False):
        self.conn: sqlite3.Connection = None
        self.cursor: sqlite3.Cursor = None
        self._in_memory: bool = in_memory
        self.database_name: str = database_name
        self.is_open = False
        if self._in_memory:
            self.file_name, self.file_folder, self.file_path = "", "", ""
            self.log_file = ""
        else:
            self.file_name, self.file_folder, self.file_path = \
                self.__generate_path_from_database_name(self.database_name)
            self.log_file = os.path.join(self.file_folder, self.file_name + self.LOG_EXTENSION)
            self.__setup_logger()
            self.logger.info(f"Database storage_provider initialization started.")
            self.open(self.file_path)
            self.logger.info(f"Database storage_provider initialization finished.")

    def __setup_logger(self):
        if self._in_memory:
            return
        self.logger = logging.getLogger("storage_provider")
        handler = logging.FileHandler(self.log_file, mode='w')
        formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.logger.setLevel(self.LOG_LEVEL)

    def delete_log_file(self):
        """Deletes the database log file."""
        try:
            if path.exists(self.log_file):
                os.remove(self.log_file)
            return True
        except OSError:
            return False

    def delete(self):
        """Closes an open database and deletes the database file."""
        if self._in_memory:
            return
        try:
            self.close()
            if path.exists(self.file_path):
                os.remove(self.file_path)
            self.logger.info(f"Database file '{self.file_path}' has been deleted.")
            return True
        except OSError as err:
            self.logger.error(f"Failed to delete database file '{self.file_path}'. {str(err)}")
            return False

    def open(self, file_path: str) -> bool:
        if self._in_memory:
            return True
        """
        Opens a database. If the database does not exist, a new database will be created.
        """
        if self.conn and (self.file_path.lower() != file_path.lower()):
            self.close()
            self.logger.info(f"Database connection to file '{self.file_path}' closed.")

        self.file_path = file_path
        self.file_folder, self.file_name = os.path.split(file_path)

        try:
            database_already_existed = Path(self.file_path).exists()
            self.conn = sqlite3.connect(self.file_path)
            self.cursor = self.conn.cursor()
            self.is_open = True
            if database_already_existed:
                self.logger.info(f"Open database file '{file_path}'.")
                return self.__validate_db()
            else:  # new database
                self.logger.info(f"Created new database file '{file_path}'.")
                self.__configure_db()
                self.__initialize_db()
                return True
        except sqlite3.Error as err:
            self.logger.error(f"Failed to open database '{file_path}'. {str(err)}")
            raise FatalException()
        except Exception as err:
            self.logger.error(f"Failed to open database '{file_path}'. {str(err)}")
            raise FatalException()

    def close(self):
        if self._in_memory:
            return
        """Closes the database."""
        if self.logger:
            self.logger.info(f"Closing database '{self.database_name}'.")
            self.logger.handlers[0].flush()
        if self.conn:
            self.commit()
            self.conn.close()
        self.is_open = False
        return True

    def commit(self, optimize=False):
        if self._in_memory:
            return
        """Commits all accumulated changes to the database."""
        if optimize:
            self.__execute('PRAGMA vacuum;')
            self.__execute('PRAGMA optimize;')
        if self.conn:
            self.conn.commit()

    def cube_get(self, cube_name, address, measure) -> float:
        """Returns a single measure from a cube fact table.
        If the idx_address does not exist, value 0.0 will be returned.
        Arguments 'idx_address' and 'measure' are index values of typ <int>."""
        if not isinstance(measure, Iterable):
            measure = [measure]
        fields_clause = ', '.join(['m' + str(m) for m in measure])
        sql = f"SELECT {fields_clause} FROM {Backend.CUB_PREFIX + cube_name} " \
              f"WHERE {' AND '.join(['d' + str(i + 1) + '=' + str(d) for i, d in enumerate(address)])};"
        # records = self.cursor.execute(sql).fetchall()
        records = self.__fetchall(sql)
        if records:
            return records[0][0]
        return 0.0

    def cube_get_range(self, cube_name, member_lists: list[list[int]], measures, aggregate: bool = True) -> list:
        """Executes a range idx_address on the cube fact table """
        if not isinstance(measures, Iterable):
            measures = [measures]
        member_list_text = []
        for i, ml in enumerate(member_lists):
            member_list_text.append(f"d{i + 1} in ({','.join([str(m) for m in ml])})")
        where_clause = ' AND '.join([m for m in member_list_text])
        if aggregate:
            fields_clause = ', '.join(['SUM(m' + str(m) + ')' for m in measures])
        else:
            fields_clause = ', '.join(['m' + str(m) for m in measures])
        sql = f"SELECT {fields_clause} " \
              f"FROM {Backend.CUB_PREFIX + cube_name} " \
              f"WHERE ({where_clause});"
        return self.__fetchall(sql)

    def cube_get_many(self, cube_name, address: list[int], measures: list[int]) -> list:
        """Returns multiple measures (given by a list of measure indexes) from a cube fact table.
        If the idx_address does not exist, a list of 0.0 values will be returned.
        Arguments 'idx_address' and 'measures' represent index values of typ list[int]."""
        sql = f"SELECT m{', '.join(['m' + str(m) for m in measures])} " \
              f"FROM {Backend.CUB_PREFIX + cube_name} " \
              f"WHERE {' AND '.join(['d' + str(i + 1) + '=' + str(d) for i, d in enumerate(address)])};"
        rs = self.__fetchall(sql)
        if rs:
            return [float(v) for v in rs[0]]
        return [0.0 for _ in measures]

    def cube_set(self, cube_name, address: list[int], measure: int, value, instant_commit: bool = True):
        """Sets single measure value into a cube fact table.
        Note: This method executes an 'upsert' on a cube fact tables."""
        table = Backend.CUB_PREFIX + cube_name
        if value is None:
            where_statement = ' AND '.join([('d' + str(i + 1) + '=' + str(d)) for i, d in enumerate(address)])
            sql = f"DELETE FROM {table} WHERE {where_statement};"
        else:
            dim_col_list = ', '.join(['d' + str(i + 1) for i, d in enumerate(address)])
            measure_col = f"m{measure}"
            sql = f"INSERT INTO {table}({dim_col_list}, {measure_col})" \
                  f"VALUES({', '.join([str(d) for d in address])}, {value}) " \
                  f"ON CONFLICT({dim_col_list}) " \
                  f"DO UPDATE SET {measure_col}=EXCLUDED.{measure_col};"
        self.__execute(sql)
        if instant_commit:
            self.commit()
        return True

    def cube_set_many(self, cube_name, address, measures, values):
        """Sets multiple value for multiple measures in a cube table.
        This method executes an upsert (insert or update) on cube tables.
        Arguments 'idx_address' and 'measures' are index values of typ <int>."""
        table = Backend.CUB_PREFIX + cube_name
        dim_col_list = ', '.join(['d' + str(i + 1) for i, d in enumerate(address)])
        dim_member_list = ', '.join([str(d) for d in address])
        value_list = ', '.join([str(v) for v in values])
        measure_col_list = ', '.join(['m' + str(m) for m in measures])
        measure_col_set_list = ', '.join([f"{'m' + str(m)}=EXCLUDED.{'m' + str(m)}" for m in measures])
        sql = f"INSERT INTO {table}({dim_col_list}, {measure_col_list})" \
              f"VALUES({dim_member_list}, {value_list}) " \
              f"ON CONFLICT({dim_col_list}) " \
              f"DO UPDATE SET {measure_col_set_list};"
        self.__execute(sql)

    def add_cube(self, cube_name, dimensions, measures):
        """Initializes the database with all required meta tables."""
        if self._in_memory:
            return
        table_name = Backend.CUB_PREFIX + cube_name
        dim_col_list = ', '.join(['d' + str(i + 1) + ' int' for i, d in enumerate(dimensions)])
        dim_col_pk_list = ', '.join(['d' + str(i + 1) for i, d in enumerate(dimensions)])
        measure_list = ', '.join(['m' + str(m) + ' real' for m in measures])
        sql = f"CREATE TABLE IF NOT EXISTS {table_name} (" \
              f"{dim_col_list}, {measure_list}, " \
              f"PRIMARY KEY ({dim_col_pk_list})) WITHOUT ROWID;"
        self.__execute(sql)
        self.logger.info(f"New cube '{cube_name}' added: {sql}")
        return table_name

    def cube_remove(self, cube_name):
        table = Backend.CUB_PREFIX + cube_name
        sql = f"DROP TABLE IF EXISTS {table};"
        self.__execute(sql)
        self.logger.info(f"Cube '{cube_name}' removed: {sql}")


    def __initialize_db(self):
        """Initializes a new and empty database by adding several meta tables."""
        self.logger.info(f"Initialization of new database started.")
        self.__add_table(self.META_TABLE_CUB, self.META_TABLE_FIELDS)
        self.__add_table(self.META_TABLE_DIM, self.META_TABLE_FIELDS)
        if not self.__table_exists(self.META_TABLE_DIM):
            self.logger.error(f"Failed to add meta tables.")
            raise FatalException("Failed to add meta tables to database.")
        self.logger.info(f"Initialization of new database finished.")

    def __validate_db(self) -> bool:
        """Checks that the database is equipped with all required meta tables."""
        if not self.__table_exists(self.META_TABLE_CUB):
            self.__initialize_db()
        return True

    def __configure_db(self):
        """Configures a new SQLite database for optimized performance."""
        self.__execute('PRAGMA temp_store=MEMORY;')
        self.__execute('PRAGMA journal_mode=MEMORY;')
        # self.__execute('PRAGMA synchronous=normal;')
        # self.__execute('PRAGMA mmap_size = 30000000000;')
        self.commit()

    def __add_table(self, table_name, fields: list[tuple]):
        sql = f"DROP TABLE IF EXISTS {table_name};"
        self.__execute(sql)
        sql = f"CREATE TABLE IF NOT EXISTS {table_name} ({', '.join([str(f[0]) + ' ' + str(f[1]) for f in fields])} );"
        self.__execute(sql)
        self.commit()
        self.logger.info(f"New database table '{table_name}' added. {sql}")

    # def __generate_add_table_statement(self, table_name, fields: list[tuple]):
    #     return f"CREATE TABLE IF NOT EXISTS {table_name} ({','.join([str(f[0]) + ' ' + str(f[1]) for f in fields])} );"

    def __get_all_db_tables(self):
        return self.__fetchall(f"SELECT name, sql FROM sqlite_master WHERE type='table';")

    def __table_exists(self, table_name):
        # get the count of tables with the name
        if not self.cursor:
            self.cursor = self.conn.cursor()
        sql = f"SELECT count(name) FROM sqlite_master WHERE type='table' AND name='{table_name}'"
        try:
            return self.cursor.execute(sql).fetchone()[0] == 1
        except sqlite3.Error as err:
            raise FatalException()

    def __execute(self, sql: str, data=None):
        """Executes an SQL idx_address without returning a result or resultset."""
        duration = 0.0
        if self.LOG_LEVEL == logging.DEBUG:
            duration = timer()

        try:
            if not self.cursor:
                self.cursor = self.conn.cursor()
            if data:
                self.cursor.execute(sql, data)
            else:
                self.cursor.execute(sql)
        except sqlite3.Error as err:
            self.logger.error(f"SQLite error on '.execute(sql)'. {str(err)} SQL := {sql}")
        except Exception as err:
            self.logger.error(f"Unexpected error on '.execute(sql)'.{str(err)} SQL := {sql}")

        if self.LOG_LEVEL == logging.DEBUG:
            duration = timer() - duration
            self.logger.debug(f"SQL statement executed in {duration:.6f}s: {sql}")

    def __execute_transaction(self, sql: list[str]):
        """Executes a list of SQL statements within a single transaction with rollback on failure."""
        duration = 0.0
        if self.LOG_LEVEL == logging.DEBUG:
            duration = timer()

        current_statement = ""
        if not self.cursor:
            self.cursor = self.conn.cursor()
        try:
            # clear current dim table
            self.cursor.execute("begin")
            for statement in sql:
                current_statement = statement
                self.cursor.execute(statement)
            self.cursor.execute("commit")
        except sqlite3.Error as err:
            self.logger.error(f"SQLite error  on '.execute_transaction(sql)'. {str(err)} SQL := {current_statement}")
            self.cursor.execute("rollback")
            return False
        except Exception as err:
            self.logger.error(f"Unexpected error on '.execute_transaction()'.{str(err)} SQL := {current_statement}")
            return False

        if self.LOG_LEVEL == logging.DEBUG:
            duration = timer() - duration
            self.logger.debug(f"SQL transaction executed in {duration:.6f}s: {sql}")
        return True

    def __fetchall(self, sql: str):
        """Executes an SQL idx_address statement and returns all records as the resultset."""
        duration = 0.0
        if self.LOG_LEVEL == logging.DEBUG:
            duration = timer()

        if not self.cursor:
            self.cursor = self.conn.cursor()
        result = self.cursor.execute(sql).fetchall()

        if self.LOG_LEVEL == logging.DEBUG:
            duration = timer() - duration
            self.logger.error(f"SQL fetchall executed in {duration:.6f}s: {sql}")

        return result

    def __generate_path_from_database_name(self, database_name: str):
        if self._in_memory:
            return None, None, None
        file_name = database_name + self.DB_EXTENSION
        file_folder = os.path.join(os.getcwd(), self.DB_FOLDER_NAME)
        if not path.isdir(file_folder):
            os.mkdir(file_folder)
        file_path = os.path.join(file_folder, file_name)
        return file_name, file_folder, file_path

    def dimension_update(self, dimension, json):
        if self._in_memory:
            return
        """"""
        data = [dimension.name, json]
        sql = f"INSERT INTO {self.META_TABLE_DIM} (key, config)" \
              f"VALUES(?, ?)  " \
              f"ON CONFLICT(key) DO UPDATE SET config = excluded.config;"
        self.__execute(sql, data)
        self.commit()
        self.logger.info(f"Configuration of dimension '{dimension.name}' added or updated in meta table.")

    def dimension_remove(self, dimension):
        if self._in_memory:
            return
        """Deletes all tables related to a dimension."""
        self.__execute(f"DELETE FROM {self.META_TABLE_DIM} WHERE key = '{dimension.name}';")
        self.logger.info(f"Dimension '{dimension.name}' removed from meta table.")

    def meta_dim(self):
        """Returns all dimension meta data."""
        return self.__fetchall(f"SELECT * FROM {self.META_TABLE_DIM}")

    def dimensions_count(self):
        """Returns the number of dimensions defined in the database."""
        return self.__fetchall(f"SELECT COUNT(*) FROM {self.META_TABLE_DIM}")[0]



# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import base64
from abc import ABC, abstractmethod

import secrets
from base64 import urlsafe_b64encode as b64e, urlsafe_b64decode as b64d
from enum import Enum, IntEnum

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


class EncryptionMethodEnum(IntEnum):
    NoEnryption = 0
    Obfuscation = 1
    Encryption = 2


class Encryptor(ABC):
    @abstractmethod
    def encrypt(self, data: str) -> str:
        """
        Encrypts a string.
        :param data: The string data to be encrypted.
        :return: The encrypted data.
        """
        pass

    @abstractmethod
    def decrypt(self, encrypted_data: str) -> str:
        """
        Decrypts an encrypted string.
        :param encrypted_data: The string data to be decrypted.
        :return: The decrypted (original) data.
        """
        pass


class NotAnEncryptor(Encryptor):
    """
    WARNING: THIS IS NOT A SECURE ENCRYPTION METHOD!
    Encryption using no encryption, meaning input = output for both methods, ``encrypt()`` and ``decrypt()``.
    """

    def encrypt(self, data) -> str:
        """
        Encrypts a string.
        :param data: The string data to be encrypted.
        :return: The encrypted data.
        """
        return data

    def decrypt(self, encrypted_data: str) -> str:
        """
        Decrypts an encrypted string.
        :param encrypted_data: The string data to be decrypted.
        :return: The decrypted (original) data.
        """
        return encrypted_data


class ObfuscationEncryptor(Encryptor):
    """
    WARNING: THIS IS NOT A SECURE ENCRYPTION METHOD!
    Encryption by data obfuscation using base64-encoding and some shifting based on a password.
    Sufficient for minimal data protection.

    Please be aware that professional hackers will likely be able to crack this encryption quite fast.
    That said, please be also aware that if you change the default key and forget it, then you might
    need to hire and pay a professional hacker to decrypt your data.
    """

    def __init__(self, password: str = "poor man's encryption"):
        self.obfuscation_key = base64.urlsafe_b64encode(password.encode()).decode()

    def encrypt(self, data: str) -> str:
        """
        Encrypts a string.
        :param data: The string data to be encrypted.
        :return: The encrypted data.
        """
        enc = []
        for i in range(len(data)):
            key_c = self.obfuscation_key[i % len(self.obfuscation_key)]
            enc_c = chr((ord(data[i]) + ord(key_c)) % 256)
            enc.append(enc_c)
        return base64.urlsafe_b64encode("".join(enc).encode()).decode()

    def decrypt(self, encrypted_data: str) -> str:
        """
        Decrypts an encrypted string.
        :param encrypted_data: The string data to be decrypted.
        :return: The decrypted (original) data.
        """
        dec = []
        enc = base64.urlsafe_b64decode(encrypted_data).decode()
        for i in range(len(enc)):
            key_c = self.obfuscation_key[i % len(self.obfuscation_key)]
            dec_c = chr((256 + ord(enc[i]) - ord(key_c)) % 256)
            dec.append(dec_c)
        return "".join(dec)


class FernetEncryptor(Encryptor):
    """
    Encryption using Fernet encryption. (see https://cryptography.io/en/latest/fernet/).
    Can be treated as a quite secure encryption (as of 2021).
    """

    def __init__(self, password: str = "superman's encryption"):
        self.backend = default_backend()
        self.iterations = 10_000
        self.password = password

    def encrypt(self, data: str) -> str:
        """
        Encrypts a string.
        :param data: The string data to be encrypted.
        :return: The encrypted data.
        """
        return self._fernet_encrypt(data.encode(), self.password).decode()

    def decrypt(self, encrypted_data: str) -> str:
        """
        Decrypts an encrypted string.
        :param encrypted_data: The string data to be decrypted.
        :return: The decrypted (original) data.
        """
        return self._fernet_decrypt(encrypted_data.encode(), self.password).decode()

    def _derive_key(self, password: bytes, salt: bytes) -> bytes:
        """Derive a secret key from a given password and salt"""
        kdf = PBKDF2HMAC(algorithm=hashes.SHA256(), length=32, salt=salt,
                         iterations=self.iterations, backend=self.backend)
        return b64e(kdf.derive(password))

    def _fernet_encrypt(self, message: bytes, password: str) -> bytes:
        salt = secrets.token_bytes(16)
        key = self._derive_key(password.encode(), salt)
        return b64e(b'%b%b%b' % (
                salt,
                self.iterations.to_bytes(4, 'big'),
                b64d(Fernet(key).encrypt(message)),
            ))

    def _fernet_decrypt(self, tokens: bytes, password: str) -> bytes:
        decoded = b64d(tokens)
        salt, iter, token = decoded[:16], decoded[16:20], b64e(decoded[20:])
        iterations = int.from_bytes(iter, 'big')
        key = self._derive_key(password.encode(), salt)
        return Fernet(key).decrypt(token)



# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import collections
import collections.abc
# from inspect import isroutine, getsource, getfile, getsourcefile
import inspect
import json
from collections.abc import Iterable

from storage.storageprovider import StorageProvider
from tinyolap.area import Area
from tinyolap.case_insensitive_dict import CaseInsensitiveDict
from tinyolap.cell import Cell
from tinyolap.dimension import Dimension
from tinyolap.exceptions import *
from tinyolap.facttable import FactTable
from tinyolap.rules import Rule, Rules, RuleError, RuleScope, RuleInjectionStrategy


class Cube:
    """Represents a multi-dimensional table."""
    __magic_key = object()

    @classmethod
    def create(cls, storage_provider: StorageProvider, name: str = None,
               dimensions: list[Dimension] = None, measures: list[str] = None,
               description: str = None):
        cube = Cube(Cube.__magic_key, name, dimensions, measures, description)
        cube._storage_provider = storage_provider
        if name and dimensions:
            if storage_provider and storage_provider.connected:
                storage_provider.add_cube(name, cube.to_json())
        return cube

    def __init__(self, cub_creation_key, name: str, dimensions=None, measures=None, description: str = None):
        """
        NOT INTENDED FOR DIRECT USE! Cubes and dimensions always need to be managed by a Database.
        Use method 'Database.add_cube(...)' to create objects type Cube.

        :param name: NAme of the cube
        :param dimensions: A list of dimensions that defines the cube axis.
        :param measures: A list if measures.
        :param description: (optional) description for the cube.
        """
        assert (cub_creation_key == Cube.__magic_key), \
            "Objects of type Cube can only be created through the method 'Database.add_cube()'."

        cube_creation_from_json = bool(dimensions)

        #self.data = [1.0 for i in range(1_200_000)]
        self._name = name
        self._description = description
        if cube_creation_from_json:
            self._dim_count = len(dimensions)
            self._dimensions = tuple(dimensions)
        else:
            self._dim_count = 0
            self._dimensions = tuple()

        self._dim_names = []
        self._dim_lookup = CaseInsensitiveDict([(dim.name, idx) for idx, dim in enumerate(self._dimensions)])
        self._facts = FactTable(self._dim_count, self)

        self._database = None
        self._storage_provider: StorageProvider = None

        # new approach
        self._rules = Rules(self._name)

        # old approach
        self._rules_all_levels = Rules(self._name)
        self._rules_base_level = Rules(self._name)
        self._rules_aggr_level = Rules(self._name)
        self._rules_roll_up = Rules(self._name)
        self._rules_on_entry = Rules(self._name)

        # create a default-measure if none is defined
        if not measures:
            measures = ["value"]
        self._measures = {}
        for idx, measure in enumerate(measures):
            self._measures[measure] = idx
        self._default_measure = measures[0]

        # initialize a default-measure if none is defined

        self._cell_request_counter: int = 0
        self._rule_request_counter: int = 0
        self._aggregation_counter: int = 0
        self._caching = True
        self._cache = {}

    def __str__(self):
        return f"cube '{self.name}'"

    def __repr__(self):
        return f"cube '{self.name}'"

    # region Rules
    def remove_rule(self, pattern: list[str]) -> bool:
        """
        Removes (unregisters) a rule function from the cube.

        :param pattern: The trigger of the rule to be removed.
        :return: ``True``, if a function with the given trigger was found and removed, ``False`` otherwise.
        """
        return NotImplemented

    def remove_all_rules(self):
        """
        Removes all rule functions from the cube.
        """
        return NotImplemented

    def register_rule(self, function, trigger: list[str] = None,
                      scope: RuleScope = RuleScope.ALL_LEVELS,
                      injection: RuleInjectionStrategy = RuleInjectionStrategy.NO_INJECTION,
                      code: str = None):
        """
        Registers a rule function for the cube. Rules function either need to be decorated with the ``@rules(...)``
        decorator or the arguments ``trigger`` and ``scope`` of the ``add_rules(...)`` function must be specified.

        :param code: (optional)Source code of the function.
        :param injection: The injection strategy defined for the function.
        :param function: The rules function to be called.
        :param trigger: The cell idx_address trigger that should trigger the rule.
        :param scope: The scope of the rule.
        """

        offset = 0
        if not inspect.isroutine(function):
            if callable(function) and function.__name__ == "<lambda>":
                offset = 1
            else:
                raise RuleException(
                    f"Argument 'function' does not seem to be a Python function, type id '{type(function)}'.")

        # validate function and decorator parameters
        function_name = str(function).split(" ")[1 + offset]
        cube_name = self.name
        if hasattr(function, "cube"):
            cube_name = function.cube
            if cube_name.lower() != self.name.lower():
                raise RuleException(
                    f"Failed to add rule function. Function '{function_name}' does not seem to be associated "
                    f"with this cube '{self.name}', but with cube '{cube_name}'.")
        if not trigger:
            if hasattr(function, "trigger"):
                trigger = function.trigger
                if type(trigger) is str:
                    trigger = [trigger, ]
                if not type(trigger) is list:
                    raise RuleException(f"Failed to add rule function. Argument 'trigger' is not of the expected "
                                        f"type 'list(str)' but of type '{type(trigger)}'.")
            else:
                raise RuleException(f"Failed to add rule function. Argument 'trigger' missing for "
                                    f"function {function_name}'. Use the '@rule(...) decorator from tinyolap.decorators.")

        if not scope:
            if hasattr(function, "scope"):
                scope = function.scope
                if not (str(type(scope)) == str(type(RuleScope.ROLL_UP))):
                    raise RuleException(f"Failed to add rule function. Argument 'scope' is not of the expected "
                                        f"type ''{type(RuleScope.ALL_LEVELS)}' but of type '{type(scope)}'.")
            else:
                raise RuleException(f"Failed to add rule function. Argument 'scope' missing for "
                                    f"function {function_name}'. Use the '@rule(...) decorator from tinyolap.decorators.")
        if not injection:
            if hasattr(function, "injection"):
                injection = function.injection
            else:
                injection = RuleInjectionStrategy.NO_INJECTION

        if type(trigger) is str:  # a lazy user forgot to put the trigger in brackets
            trigger = [trigger, ]

        # setup and add rule
        idx_pattern = self._pattern_to_idx_pattern(trigger)
        rule = Rule(function=function, name= function_name, cube=self.name, trigger=trigger, idx_trigger_pattern=idx_pattern,
                    scope=scope, injection=injection, code=code)
        self._rules.add(rule)

        # old approach
        # if scope == RuleScope.ALL_LEVELS:
        #     self._rules_all_levels.register(function, self._name, function_name, trigger, idx_pattern, scope, injection, code)
        # elif scope == RuleScope.AGGREGATION_LEVEL:
        #     self._rules_aggr_level.register(function, self._name, function_name, trigger, idx_pattern, scope, injection, code)
        # elif scope == RuleScope.BASE_LEVEL:
        #     self._rules_base_level.register(function, self._name, function_name, trigger, idx_pattern, scope, injection, code)
        # elif scope == RuleScope.ROLL_UP:
        #     self._rules_roll_up.register(function, self._name, function_name, trigger, idx_pattern, scope, injection, code)
        # elif scope == RuleScope.ON_ENTRY:
        #     self._rules_on_entry.register(function, self._name, function_name, trigger, idx_pattern, scope, injection, code)
        # else:
        #     raise RuleException(f"Unexpected value '{str(scope)}' for argument 'scope: RuleScope'.")

        # add function to code manager
        self._database._code_manager.register_function(
            function=function, cube=cube_name, trigger=trigger,
            scope=scope, injection=injection, code=code)

    def validate_rules(self, save_after_sucessfull_validation: bool = True) -> (bool, str):
        """
        Validates all registered rules by calling each with a random cell matching the defined
        rule trigger and rule scope. Calling this methods (maybe even multiple times) can be
        usefull for highlevel testing of your rules.

        .. warning::
            Calling this method does not replace proper rule testing. Escpecially when your
            database should be used by other users or for serious purposes you need to ensure
            that your rule calculations ideally never (or only for explicit purposes) thows an
            error.

        :param save_after_sucessfull_validation: If set to true
        :return: ``(True, validation_results_json:str)`` if all rules returned a proper result without causing errors.
                 ``(False, validation_results_json:str)`` if at least one rule causes an error. The validation results
                  will contain information for all rules that have been processed. The validation will not stop
                  on the first error casued by a rule function.
        """
        # todo: to be implemented

        # update the database
        if save_after_sucessfull_validation:
            self._database.save()
        return True

    def _pattern_to_idx_pattern(self, pattern):
        """
        Converts a trigger into it's index representation.

        :param pattern: The trigger to be converted.
        :return: The index trigger.
        """
        if type(pattern) is str:
            pattern = list((pattern,))
        # Sorry, miss-use of cursor. maybe some refactoring required...
        address = self._get_default_cell_address()
        c = self._create_cell_from_bolt(address, self._address_to_bolt(address))
        # create something like this: idx_pattern = [(0, 3)]
        idx_pattern = []
        for p in pattern:
            idx_dim, idx_member, member_level = c._get_member(p)
            idx_pattern.append((idx_dim, idx_member))
        return idx_pattern
    # endregion

    # region Properties
    @property
    def cells_count(self) -> int:
        return len(self._facts.facts)

    @property
    def name(self) -> str:
        """Returns the name of the cube."""
        return self._name

    @property
    def counter_cell_requests(self) -> int:
        """Returns the number of cell requests executed."""
        return self._cell_request_counter

    @property
    def counter_rule_requests(self) -> int:
        """Returns the number of rules that have been executed."""
        return self._rule_request_counter

    @property
    def counter_aggregations(self) -> int:
        """Returns the number aggregations that have been executed."""
        return self._aggregation_counter

    def reset_counters(self):
        """Resets the internal counters for cell- and rule-requests and aggregations."""
        self._cell_request_counter = 0
        self._rule_request_counter = 0
        self._aggregation_counter = 0

    @property
    def caching(self) -> bool:
        """Identifies if caching is activated for the current cube.
        By default, caching is activated for all cubes."""
        return self._caching

    @caching.setter
    def caching(self, value: bool):
        """Identifies if caching is activated for the current cube.
        By default, caching is activated for all cubes."""
        self._caching = value
        if not self._caching:
            self._cache = {}

    @property
    def dimensions_count(self) -> int:
        """Returns the number of dimensions of the cube."""
        return self._dim_count

    @property
    def measures_count(self) -> int:
        """Returns the number of measures of the cube."""
        return len(self._measures)

    @property
    def default_measure(self) -> str:
        """Returns/sets the default measure of the cube.
        By default, this is always the first measure from the list of measures."""
        return self._default_measure

    @default_measure.setter
    def default_measure(self, value: str):
        """Returns/sets the default measure of the cube.
        By default, this is always the first measure from the list of measures."""
        if value not in self._measures:
            raise KeyNotFoundError(f"Failed to set default member. "
                                   f"'{value}' is not a measure of cube '{self._name}'.")
        self._default_measure = value

    @property
    def measures(self) -> list[str]:
        """Returns the list of measures of a cube."""
        return [str(self._measures.keys())]

    # endregion

    # region Dimension and Measures related methods
    def get_measures(self) -> list[str]:
        """Returns the list of measures of a cube."""
        return list(self._measures.keys())

    def get_dimension_by_index(self, index: int):
        """Returns a dimension from a cubes list of dimensions at the given index."""
        if (index < 0) or (index > self._dim_count):
            raise ValueError(f"Requested dimension index '{index}' is out of range [{0}, {self._dim_count}].")
        return self._dimensions[index]

    def get_dimension_ordinal(self, name: str):
        """Returns the ordinal position of a dimension with the cube definition.
        :return The ordinal position of the dimension, if the dimension is contained ones in the cube.
                A list of ordinal positions if the dimension is contained multiple times in the cube.
        """
        ordinals = []
        for idx, dim_name in enumerate([dim.name for dim in self._dimensions]):
            if name == dim_name:
                ordinals.append(idx)
        if not ordinals:
            return -1
        if len(ordinals) == 1:
            return ordinals[0]
        return ordinals

    def get_dimension(self, name: str):
        """Returns the dimension defined for the given dimension index."""
        result = [dim for dim in self._dimensions if dim.name == name]
        if not result:
            raise ValueError(f"Requested dimension '{name}' is not a dimension of cube {self.name}.")
        return result[0]

    # endregion

    # region Cell access via indexing/slicing
    def __getitem__(self, item):
        if len(item) == self._dim_count:
            bolt = self._address_to_bolt(item)
            return self._get(bolt)
        else:
            return self.area(item)

    def __setitem__(self, item, value):
        if len(item) == self._dim_count:
            bolt = self._address_to_bolt(item)
            self._set(bolt, value)
        else:
            self.area(item).set_value(value)

    def __delitem__(self, item):
        if len(item) == self._dim_count:
            bolt = self._address_to_bolt(item)
            self._set(bolt, None)
        else:
            self.area(item).clear()

    # endregion

    # region Read and write values
    def clear(self):
        """
        Removes all values from the Cube.
        """
        self._facts.clear()

    def get(self, address: tuple):
        """Reads a value from the cube for a given idx_address.
        If no records exist for the given idx_address, then 0.0 will be returned.
        :raises InvalidKeyException:
        """
        bolt = self._address_to_bolt(address)
        return self._get(bolt)

    def set(self, address: tuple, value):
        """Writes a value to the cube for the given bolt (idx_address and measures)."""
        bolt = self._address_to_bolt(address)
        return self._set(bolt, value)

    def _get(self, bolt, bypass_rules=False):
        """
        Returns a value from the cube for a given idx_address and measure.
        If no records exist for the given idx_address, then 0.0 will be returned.
        """
        (super_level, idx_address, idx_measures) = bolt
        self._cell_request_counter += 1

        # caching
        if not bypass_rules:
            if self._caching and bolt in self._cache:
                return self._cache[bolt]

        # ALL_LEVELS rules
        if not bypass_rules:
            found, func = self._rules.match(scope=RuleScope.ALL_LEVELS, idx_address=idx_address)
            if found:
                cursor = self._create_cell_from_bolt(None, (super_level, idx_address, idx_measures))
                try:
                    self._rule_request_counter += 1
                    value = func(cursor)
                    if value != Cell.CONTINUE:
                        if self._caching:
                            self._cache[bolt] = value  # save value to cache
                        return value
                except ZeroDivisionError:
                    return RuleError.DIV0
                except Exception as err:
                    return RuleError.ERROR

        if super_level == 0:  # base-level cells
            # BASE_LEVEL rules
            if not bypass_rules:
                found, func = self._rules.match(scope=RuleScope.BASE_LEVEL, idx_address=idx_address)
                if found:
                    cursor = self._create_cell_from_bolt(None, (super_level, idx_address, idx_measures))
                    try:
                        self._rule_request_counter += 1
                        value = func(cursor)
                        if value != Cell.CONTINUE:
                            if self._caching:
                                self._cache[bolt] = value  # save value to cache
                            return value
                    except ZeroDivisionError:
                        return RuleError.DIV0
                    except Exception as err:
                        return RuleError.ERROR

            if type(idx_measures) is int:
                return self._facts.get(idx_address, idx_measures)
            else:
                raise FatalException("Depreciated. Feature Needs to be removed")

        else:  # aggregated cells
            # AGGREGATION_LEVEL
            if not bypass_rules:
                found, func = self._rules.match(scope=RuleScope.AGGREGATION_LEVEL, idx_address=idx_address)
                if found:
                    cursor = self._create_cell_from_bolt(None, (super_level, idx_address, idx_measures))
                    try:
                        self._rule_request_counter += 1
                        value = func(cursor)
                        if value != Cell.CONTINUE:
                            if self._caching:
                                self._cache[bolt] = value  # save value to cache
                            return value
                    except ZeroDivisionError:
                        return RuleError.DIV0
                    except Exception as err:
                        return RuleError.ERROR

            # get records row ids for current cell idx_address
            rows = self._facts.query(idx_address)
            self._aggregation_counter += len(rows)

            # aggregate records
            if type(idx_measures) is int:
                if not rows:
                    return 0.0
                facts = self._facts.facts
                total = 0.0

                # todo: add support for ROLL_UP rules

                for row in rows:
                    if idx_measures in facts[row]:
                        value = facts[row][idx_measures]
                        if type(value) is float:
                            total += value

                if self._caching:
                    self._cache[bolt] = total  # save value to cache

                return total
            else:
                raise FatalException("Depreciated. Feature Needs to be removed")

    def _set_base_level_cell(self, idx_address, idx_measure, value):
        """Writes a base level value to the cube for the given idx_address (idx_address and measures)."""
        self._facts.set(idx_address, idx_measure, value)
        if self._storage_provider and self._storage_provider.connected:
            if value:
                self._storage_provider.set_record(self._name, str(idx_address), json.dumps({"value": value}))
            else:
                self._storage_provider.set_record(self._name, str(idx_address))

    def _set(self, bolt, value, bypass_rules=False):
        """Writes a value to the cube for the given bolt (idx_address and measures)."""
        if self._caching and self._cache:
            self._cache = {}  # clear the cache

        (super_level, idx_address, idx_measures) = bolt

        if type(value) is int:
            value = float(value)

        if super_level == 0:  # for base-level cells...
            if type(idx_measures) is int:
                result = self._facts.set(idx_address, idx_measures, value)
                if self._storage_provider and self._storage_provider.connected:
                    if value:
                        self._storage_provider.set_record(self._name, str(idx_address), json.dumps({"value": value}))
                    else:
                        self._storage_provider.set_record(self._name, str(idx_address))

            #  ...check for base-level on_entry rule (aka push-rules) to be executed
            if not bypass_rules:
                found, func = self._rules.match(scope=RuleScope.ON_ENTRY, idx_address=idx_address)
                if found:
                    cursor = self._create_cell_from_bolt(None, (super_level, idx_address, idx_measures))
                    try:
                        self._rule_request_counter += 1
                        func(cursor)
                    except Exception as err:
                       pass

        else:
            raise InvalidOperationException(f"Write back to aggregated cells in not (yet) supported.")

    def _update_aggregation_index(self, fact_table_index, address, row):
        """Updates all fact table index for all aggregations over all dimensions. FOR INTERNAL USE ONLY!"""
        # please note that a '__' name prefix is not possible
        # as this function is called through a weak reference.
        for d, idx_member in enumerate(address):
            for idx_parent in self._dimensions[d].members[address[d]][self._dimensions[d].ALL_PARENTS]:
                if idx_parent in fact_table_index._index[d]:
                    fact_table_index._index[d][idx_parent].add(row)
                else:
                    fact_table_index._index[d][idx_parent] = {row}

    def _validate_address(self, address: tuple, measure):
        """Validates a given idx_address and measures and return the according indexes."""
        if type(measure) is str:
            if measure not in self._measures:
                raise ValueError(f"'{measure}' is not a measure of cube '{self.name}'.")
            idx_measure = self._measures[measure]
        elif isinstance(measure, Iterable):
            idx_measure = []
            for m in measure:
                if m not in self._measures.keys():
                    raise KeyNotFoundError(f"'{m}' is not a measure of cube '{self._name}'.")
                idx_measure.append(self._measures[m])
        else:
            idx_measure = self._measures[self._default_measure]

        if len(address) != self._dim_count:
            raise ValueError("Invalid number of dimensions in idx_address.")
        idx_address = list(range(0, self._dim_count))
        super_level = 0
        for d in range(0, self._dim_count):
            if address[d] in self._dimensions[d]._member_idx_lookup:
                idx_address[d] = self._dimensions[d]._member_idx_lookup[address[d]]
                super_level += self._dimensions[d].members[idx_address[d]][self._dimensions[d].LEVEL]
            else:
                raise ValueError(f"'{address[d]}' is not a member of dimension '{self._dimensions[d]._name}'.")
        return tuple(idx_address), super_level, idx_measure

    def _remove_members(self, dimension, members):
        """Removes members from indexes and data table"""
        ordinal = self.get_dimension_ordinal(dimension.name)
        if ordinal == -1:  # dimension is not contained in this cube
            return

        if type(ordinal) is int:
            ordinal = [ordinal]

        # clear fact table
        for o in ordinal:
            self._facts.remove_members(o, members)

    def _idx_address_to_address(self, idx_address, include_cube_name: bool = False):
        """
        Converts an idx_address index to a idx_address with member names
        :param idx_address:
        :return:
        """
        if include_cube_name:
            address = [self.name, ]
        else:
            address = []
        for i in range(self._dim_count):
            address.append(self._dimensions[i].members[idx_address[i]][1])
        return address

    def _address_to_idx_address(self, address):
        """
        Converts an address to an idx_address index with member ids.
        :param idx_address:
        :return:
        """
        bolt = self._address_to_bolt(address)
        return bolt[1]

    def _address_to_bolt(self, address):
        """Converts a given address, incl. member and (optional) measures, into a bolt.
        A bolt is a tuple of integer address, used for internal access of cells.
        """

        dim_count = self._dim_count
        measures_count = len(address) - dim_count
        if measures_count < 0:
            raise InvalidCellAddressException(f"Invalid idx_address. At least {self._dim_count} members expected "
                                              f"for cube '{self._name}, but only {len(address)} where passed in.")
        # Validate members
        dimensions = self._dimensions
        idx_address = [None] * dim_count
        super_level = 0
        for i, member in enumerate(address[: dim_count]):
            if member in dimensions[i]._member_idx_lookup:
                idx_address[i] = dimensions[i]._member_idx_lookup[member]
                super_level += dimensions[i].members[idx_address[i]][6]
            else:
                raise InvalidCellAddressException(f"Invalid idx_address. '{member}' is not a member of the {i}. "
                                                  f"dimension '{dimensions[i].name}' in cube {self._name}.")
        idx_address = tuple(idx_address)

        # validate measures (if defined)
        if measures_count == 0:
            idx_measures = self._measures[self._default_measure]
        else:
            idx_measures = []
            for measure in address[self._dim_count:]:
                if measure not in self._measures:
                    raise InvalidCellAddressException(f"'{measure}' is not a measure of cube '{self.name}'.")
                idx_measures.append(self._measures[measure])
            if measures_count == 1:
                idx_measures = idx_measures[0]
            else:
                idx_measures = tuple(idx_measures)

        return super_level, idx_address, idx_measures  # that's the 'bolt'

    # endregion

    # region cells
    def cell(self, *args) -> Cell:
        """Returns a new Cell from the Cube."""
        return Cell.create(self, self._dim_lookup, args, self._address_to_bolt(args))

    def _create_cell_from_bolt(self, address, bolt) -> Cell:
        """Create a Cell for the Cube directly from an existing bolt."""
        return Cell.create(self, self._dim_lookup, address, bolt)

    def _get_default_cell_address(self):
        """Generates a default address. This is the first member from all dimensions."""
        address = []
        for dim in self._dimensions:
            # keys = list(dim._member_idx_lookup.keys())
            # address.append(keys[0])
            address.append(dim.get_first_member())
        return tuple(address)

    def _get_records(self):
        """
        Returns all records and values from the cube as a list[tuple[address:str, value_as_json:str].
        Useful for database serialization.
        """
        records = []
        for address, data in zip(self._facts.addresses, self._facts.facts):
            records.append((str(address), json.dumps(data)))
        return records

    # endregion

    # region serialization
    # todo: adjust to fully support cube (e.g. rules)
    def to_json(self):
        """
        Returns the json representation of the cube. Helpful for serialization
        and deserialization of cubes. The json returned by this function is
        the same as the one used by storage providers (if available).

        :param beautify: Identifies if the json code should be beautified (multiple rows + indentation).
        :return: A json string representing the cube.
        """
        dim_names = [dim.name for dim in self._dimensions]
        rules_count = len(self._rules_roll_up) + len(self._rules_on_entry) \
                      + len(self._rules_all_levels) + len(self._rules_base_level) + len(self._rules_aggr_level)
        config = {"content": "cube",
                  "name": self.name,
                  "description": self._description,
                  "dimensions": dim_names,
                  "caching": self._caching,
                  "rules": rules_count,
                  }

        json_string = json.dumps(config, indent=4)
        return json_string

    # todo: adjust to fully support cube (e.g. rules)
    def from_json(self, json_string: str):
        """
        Initializes the cube from a json string.

        .. warning::
            Calling this method for cubes which are already in use (contain data)
            will very likely **corrupt your database!** Calling this method is only save
            **before** you write any data to a cube. Handle with care.

        :param json_string: The json string containing the cube definition.
        :raises FatalException: Raised if an error occurred during the deserialization from json string.
        """
        try:
            # read configuration
            config = json.loads(json_string)
            self._name = config["name"]
            self._description = config["description"]
            new_dim_names = config["dimensions"]
            self._dimensions = tuple([self._database.dimensions[dim_name] for dim_name in new_dim_names])
            self._caching = config["caching"]

            self._dim_count = len(self._dimensions)
            self._dim_names = []
            self._dim_lookup = CaseInsensitiveDict([(dim.name, idx) for idx, dim in enumerate(self._dimensions)])
            self._facts = FactTable(self._dim_count, self)

            # load data
            if self._storage_provider:
                records = self._storage_provider.get_records(self._name)
                for record in records:
                    address = str(record[0])
                    idx_address = tuple(map(int, address[1:-1].split(sep=",")))

                    data = json.loads(record[1])
                    for key, value in data.items():
                        idx_measure = self._measures[key]
                        self._set_base_level_cell(idx_address, idx_measure, value)

            # initialize rules
            # Note: This should to be done after loading the data, as otherwise push rules
            #       might be triggered and recalculated although the data is (should be) already consistent.
            if config["rules"] > 0:
                pass
            functions = self._database._code_manager.get_functions(self._name)
            for f in functions:
                self.register_rule(function=f.function, trigger=f.trigger,
                                   scope= f.scope, injection=f.injection, code=f._code)

        except Exception as err:
            raise FatalException(f"Failed to load json for dimension '{self.name}'. {str(err)}")

    # endregion

    # region areas
    def area(self, *args) -> Area:
        """Returns a new Area fom the Cube."""
        return Area(self, args)

    # endregion



# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

class FactTable:
    """
    Stores all records in simple row store and maintains an index
    over all members for faster aggregations and queries.
    """

    class FactTableIndex:
        """
        Represents a minimalistic multidimensional index over a row store.
        For each member of each dimension a set is maintained containing
        the indexes of all records related to each specific member.
        """

        def __init__(self, dimensions: int):
            self.dims = dimensions
            self._index = []
            for i in range(0, self.dims):
                self._index.append({})

        def set(self, address, row):
            for i in range(0, self.dims):
                if address[i] in self._index[i]:
                    self._index[i][address[i]].add(row)
                else:
                    self._index[i][address[i]] = {row}

        def exists(self, dimension: int, key):
            return key in self._index[dimension]

        def clear(self):
            for i in range(0, self.dims):
                for member_idx, row_list in self._index[i].items():
                    self._index[i][member_idx] = set()

        def get_rows(self, dimension: int, key):
            if key in self._index[dimension]:
                return self._index[dimension][key]
            return set()

        def remove_members(self, dim_idx, members, deletes, delete_set, shifts):
            shift_set = set(shifts.keys())

            # remove the members from the index
            for member in members:
                if member in self._index[dim_idx]:
                    del self._index[dim_idx][member]

            # delete all row references of rows to be deleted
            for i in range(0, self.dims):
                for member_idx, row_list in self._index[i].items():
                    new_row_set = row_list.difference(delete_set)

                    # execute all shifts
                    shifters = new_row_set.intersection(shift_set)
                    if shifters:
                        for shifter in shifters:
                            new_row_set.remove(shifter)
                            new_row_set.add(shifts[shifter])

                    self._index[i][member_idx] = new_row_set

        def remove_rows(self, deletes, delete_set, shifts):
            shift_set = set(shifts.keys())
            # delete all row references of rows to be deleted
            for i in range(0, self.dims):
                for member_idx, row_list in self._index[i].items():
                    new_row_set = row_list.difference(delete_set)

                    # execute all shifts
                    shifters = new_row_set.intersection(shift_set)
                    if shifters:
                        for shifter in shifters:
                            new_row_set.remove(shifter)
                            new_row_set.add(shifts[shifter])

                    self._index[i][member_idx] = new_row_set
        def get_size(self) -> int:
            size = 0
            for i in range(0, self.dims):
                for m in self._index[i].values():
                    if m:
                        size += len(m)
            return size

        def get_count(self) -> int:
            count = 0
            for i in range(0, self.dims):
                count += len(self._index[i].keys())
            return count

    def __init__(self, dimensions: int, cube):
        self.row_lookup = {}
        self.facts = []
        self.addresses = []
        self.index = FactTable.FactTableIndex(dimensions)
        self.dims = dimensions
        # we need access to the parent cube. This is a garbage collection safe approach
        self.cube = cube

        self.cached_set = None
        self.cached_idx = []
        self.cached_seq = []
        self.cache_machtes: int = 0
        self.cached_seq_machting = None

    def set(self, address: tuple, measure, value):
        record_hash = hash(address)
        if record_hash in self.row_lookup:
            # overwrite existing record/value
            row = self.row_lookup[record_hash]
            self.facts[row][measure] = value
            return True
        else:
            # add new record
            row = len(self.facts)
            self.row_lookup[record_hash] = row
            self.facts.append({measure: value})
            self.addresses.append(address)
            # update index
            self.index.set(address, row)
            if self.cube:
                self.cube._update_aggregation_index(self.index, address, row)
            return True

    def get(self, address, measure):
        record_hash = hash(address)
        if record_hash in self.row_lookup:
            # overwrite existing record/value
            row = self.row_lookup[record_hash]
            if measure in self.facts[row]:
                return self.facts[row][measure]
        return 0.0

    def get_facts(self, record):
        record_hash = hash(record)
        if record_hash in self.row_lookup:
            # overwrite existing record/value
            row = self.row_lookup[record_hash]
            return self.facts[row]
        return None

    def clear(self):
        """
        Clears the fact table and resets the index.
        """
        self.row_lookup = {}
        self.facts = []
        self.addresses = []
        self.index.clear()

    def __len__(self):
        return len(self.facts)

    def get_record_by_row(self, row: int):
        return self.addresses[row], self.facts[row]

    def get_value_by_row(self, row, measure):
        if measure in self.facts[row]:
            return self.facts[row][measure]
        return None

    def query(self, idx_address):
        """
        Evaluates the record rows required for an OLAP aggregation.
        :param idx_address: The cell address to be evaluated.
        """
        get_rows = self.index.get_rows
        # get the row sets to be intersected
        sets = []
        for i in range(0, len(idx_address)):
            if idx_address[i] != 0:
                if self.index.exists(i, idx_address[i]):
                    sets.append(get_rows(i, idx_address[i]))
                else:
                    # if the key is not available in the index then no records exist, we're done
                    return set()  # an empty set

        # Order matters most!!! So we order the sets ascending by their length, to intersect smaller sets first.
        # This improves the performance of intersections quite nicely, -10% worst, +200% on average and +600% best.
        seq = sorted(((len(s), d) for d, s in enumerate(sets)))
        # Execute intersection of sets
        result = sets[seq[0][1]]
        for i in range(1, len(seq)):  # do not change to something like <for i,d in seq:>, it's slower.
            result = result.intersection(sets[seq[i][1]])
            if not result:
                # if the set is empty, then there are no records matching the requested address
                break
        return result

    def cached_query(self, idx_address):
        """
        Evaluates the records required for OLAP aggregation using a cache to minimize set operations.
        Using this method only pays back for very special use cases where many aggregations need to be performed
        and subsequent calls have very similar pattern.
        :param idx_address: The cell address to be evaluated.
        """
        get_rows = self.index.get_rows
        sets = []  # an empty set

        # check if we can use the cached base set
        base_set, seq_matching, seq_remaining, caching_requested = self._get_base_set_from_cache(idx_address)

        if base_set:
            sets.append(base_set)
            # get remaining sets to be intersected
            for d, idx in seq_remaining:
                if idx_address[d] != 0:  # 0 = "*", meaning all rows for that dimension, no processing required
                    if self.index.exists(d, idx):
                        sets.append(get_rows(d, idx))

        elif caching_requested:
            # get matching set first and evaluate those
            for d, idx in seq_matching:
                if idx_address[d] != 0:
                    if self.index.exists(d, idx):
                        sets.append(get_rows(d, idx))
                    else:
                        # if the key is not available in the index then no records exist, we're done
                        return set([])  # an empty set

            # Execute intersection of matching sets
            # Order matters most!!! So we order the sets ascending by number of items, to intersect smaller sets first,
            # this improves the performance of intersection quite nicely (±2x up to ±4x times on average).
            seq = sorted(((len(s), d) for d, s in enumerate(sets)))
            base_set = sets[seq[0][1]]
            for len_set, d in seq:
                base_set = base_set.intersection(sets[d])
            sets = [base_set]

            # save to cache
            self.cached_set = base_set
            self.cached_idx = idx_address
            self.cached_seq_machting = seq_matching

            # get remaining sets to be intersected
            for d, idx in seq_remaining:
                if idx_address[d] != 0:  # 0 = "*", meaning all rows for that dimension, no processing required
                    if self.index.exists(d, idx):
                        sets.append(get_rows(d, idx))

        else:
            # get sets to be intersected
            for i in range(0, len(idx_address)):
                if idx_address[i] != 0:
                    if self.index.exists(i, idx_address[i]):
                        sets.append(get_rows(i, idx_address[i]))
                    else:
                        # if the key is not available in the index then no records exist, we're done
                        return set([])  # an empty set
            if not sets:
                # todo: This is not an error! return all rows instead
                raise ValueError(f"Invalid query {idx_address}. At least one dimension needs to be specified.")

        seq = sorted(((len(s), d) for d, s in enumerate(sets)))
        result = sets[seq[0][1]]
        for i in range(1, len(seq)):
            result = result.intersection(sets[seq[i][1]])
            if not result:
                # if the set is empty, then there are no records matching the requested address
                break
        return result

    def _get_base_set_from_cache(self, idx_address) -> (set, list, bool):
        """checks whether a cached base set is available or must be updated."""
        if not self.cached_idx:
            # nothing cached yet
            self.cached_idx = idx_address
            return None, None, None, False

        # compare requested index with cached index
        seq_matching = []
        seq_remaining = []
        matches = 0
        for d in range(self.dims):
            if idx_address[d] == self.cached_idx[d]:
                matches += 1
                seq_matching.append((d, idx_address[d]))
            else:
                seq_remaining.append((d, idx_address[d]))

        # only if there are at least 2 matching indexes, then using the chached set will be beneficial
        if matches > 1:
            if self.cached_seq_machting and self.cached_seq_machting == seq_matching:  # ensure the cache exactly mathes
                return self.cached_set, seq_matching, seq_remaining, False  # 'False' indicates that we keep the cache
            else:
                return None, seq_matching, seq_remaining, True  # the 'True' indicates that caching this cell is requested.
        else:
            # flush the cached set, but remember the index
            self.cached_set = None
            self.cached_seq_machting = None
            self.cached_idx = idx_address
            return None, None, None, False

    def query_area(self, idx_area_def):
        sets = []  # an empty set
        # get first relevant set
        get_rows = self.index.get_rows
        for i in range(0, len(idx_area_def)):
            if idx_area_def[i]:
                for idx_member in idx_area_def[i]:
                    if idx_member != 0:  # "*" means all rows for that dimension, no processing required
                        if self.index.exists(i, idx_member):
                            sets.append(get_rows(i, idx_member))

        if not sets:
            return set()
            # todo: This is not an error! return all rows instead
            raise ValueError(f"Invalid query_area. At least one dimension needs to be specified.")

        # Execute intersection of sets
        # Order matters most!!! order the sets by ascending number of items, this greatly
        # improves the performance of intersection operations.
        seq = sorted(((len(s), i) for i, s in enumerate(sets)))
        result = sets[seq[0][1]]
        for i in range(1, len(seq)):
            result = result.intersection(sets[seq[i][1]])
        return result

    def remove_members(self, dim_idx: int, members: ()):
        """
        Removes members for a specific dimension from the fact table.
        :param dim_idx: Ordinal index of the dimension affected
        :param members: List of indexes of the members to be removed.
        :return:
        """
        # 1. find effected records
        deletes = []
        for row, address in enumerate(self.addresses):
            if address[dim_idx] in members:
                deletes.append(row)
        delete_set = set(deletes)
        # deletes[] now contains the rows to be deleted in asc sorted order.

        # 2. prepare shifting of row positions for all records
        #    create tuples with (old_position, new position)
        shifts = {}
        shift = 0
        row = -1
        for del_row in deletes:
            if row == -1:
                shift += 1
                row = del_row + 1
            else:
                for r in range(row, del_row):  # from the start...
                    shifts[r] = r - shift
                shift += 1
                row = del_row + 1
        for r in range(row, len(self.row_lookup)):  # ...up to the end
            shifts[r] = r - shift

        # 3.1 update the lookup table indexes...
        temp_lookup = self.row_lookup.copy()
        # 3.2 ...then delete the obsolete records
        self.row_lookup = {address: row for address, row in self.row_lookup.items() if row not in delete_set}
        for k, v in temp_lookup.items():
            if v in shifts:
                self.row_lookup[k] = shifts[v]

        # 4. remove records
        self.facts = [data for row, data in enumerate(self.facts) if row not in delete_set]
        self.addresses = [data for row, data in enumerate(self.addresses) if row not in delete_set]

        # 5. finally update the index
        self.index.remove_members(dim_idx, members, deletes, delete_set, shifts)

    def remove_records(self, records: set):
        """
        Removes a list of records from the fact table.
        :param records: An iterable of int, identifying the row numbers to be removed.
        :return:
        """
        # 1. find effected records
        delete_set = records
        deletes = []
        delete_addresses = []
        for row, address in enumerate(self.addresses):
            if row in delete_set:
                deletes.append(row)
                delete_addresses.append(address)
        # deletes[] now contains the rows to be deleted in asc sorted order.
        # deletes_addresses[] contains the associated cell addresses

        # 2. prepare shifting of row positions for all records
        #    create tuples with (old_position, new position)
        shifts = {}
        shift = 0
        row = -1
        for del_row in deletes:
            if row == -1:
                shift += 1
                row = del_row + 1
            else:
                for r in range(row, del_row):  # from the start...
                    shifts[r] = r - shift
                shift += 1
                row = del_row + 1
        for r in range(row, len(self.row_lookup)):  # ...up to the end
            shifts[r] = r - shift

        # 3.1 update the lookup table indexes...
        temp_lookup = self.row_lookup.copy()
        # 3.2 ...then delete the obsolete records
        self.row_lookup = {address: row for address, row in self.row_lookup.items() if row not in delete_set}
        for k, v in temp_lookup.items():
            if v in shifts:
                self.row_lookup[k] = shifts[v]

        # 4. remove records
        self.facts = [data for row, data in enumerate(self.facts) if row not in delete_set]
        self.addresses = [data for row, data in enumerate(self.addresses) if row not in delete_set]

        # 5. finally update the index
        self.index.remove_rows(deletes, delete_set, shifts)


# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

class TinyOlapException(Exception):
    """Base class for all tinyOlap specific exceptions."""
    pass


class EncryptionException(TinyOlapException):
    """Raised when an encryption or decryption error occured. Especially on wrong password."""
    pass


class RuleException(TinyOlapException):
    """Raised when an rule function fails or is invalid."""
    pass


class InvalidCellAddressException(TinyOlapException):
    """Raised when an invalid cube cell idx_address is detected."""
    pass


class InvalidOperationException(TinyOlapException):
    """Raised when the invalid operation on a database object is executed."""
    pass


class CubeCreationException(TinyOlapException):
    """Raised when the creation of a cube."""
    pass


class DimensionInUseException(TinyOlapException):
    """Raised when the deletion of a dimension failed to to being used by a cube."""
    pass


class DatabaseBackendException(TinyOlapException):
    """Raised when an error occurs while accessing or handling database storage_provider or files."""
    pass


class FatalException(TinyOlapException):
    """Raised when a fatal error occurs."""
    pass


class KeyNotFoundError(TinyOlapException):
    """Raised when a certain member, dimension, table etc. not exist."""
    pass


class InvalidKeyException(TinyOlapException):
    """Raised when a certain name of a member, dimension, table etc. is invalid  or not supported."""
    pass


class DuplicateKeyException(TinyOlapException):
    """Raised when a certain member, dimension, table etc. was added that already exists."""
    pass


class InvalidMemberNameError(TinyOlapException):
    """Raised when a dimension member name is invalid, containing unsupported special characters."""
    pass


class DimensionEditModeException(TinyOlapException):
    """Raised when an error occurred while a dimension is in edit mode,
    or if member were added, removed or renamed while not in edit mode.
    Call 'edit_begin()' to start editing."""
    pass



#!/usr/bin/env python
from setuptools import setup

# ...to run the build and deploy process to PyPi.org:
# python3 setup.py sdist bdist_wheel   # note: Wheel need to be installed: pip install wheel
# twine upload dist/*                  # note: Twine need to be installed pip install twine

# ... via Github actions
# https://packaging.python.org/en/latest/guides/publishing-package-distribution-releases-using-github-actions-ci-cd-workflows/

VERSION = '0.8.7'
DESCRIPTION = "TinyOlap: A multi-dimensional in-memory OLAP database in plain Python 3."
LONG_DESCRIPTION = """
TinyOlap is a minimal in-process in-memory multi-dimensional database with numerical aggregations 
and calculations in mind. First a multi-dimensional data model needs to be defined, consisting of 
cubes, dimensions, members, hierarchies etc. Afterwards additional calculation logic can be added 
through arbitrary Python code. Data access is cell-based or range-based. A minimal support for SQL 
in also provided. All calculations will be executed on the fly. Optionally, persistence is provided
through SQLite. TinyOlap is a byproduct of a research project, intended to mimic the behavior and 
capabilities of real-world MOLAP databases (e.g. IBM TM/1, SAP HANA or Jedox PALO) but with a super 
minimal footprint. TinyOlap is best suited for interactive planning, forecasting, simulation and 
general multidimensional numerical problems.          
"""

setup(
    name="tinyolap",
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Utilities",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    author="Thomas Zeutschler",
    keywords=['database', 'olap', 'molap', 'planning', 'simulation', 'forecasting',
              'multidimensional', 'cube', 'business rules', 'calculation'],
    author_email="margins.roadmap_0t@icloud.com",
    url="http://github.com/zeutschler/tinyolap/",
    license="MIT License",
    platforms=['any'],
    zip_safe=True,
    python_requires='>=3.8',
    install_requires=[
        'cryptography',
        'sqlparse',
        'enum_tools'
    ],
    test_suite="tinyolap.tests",
    packages=['tinyolap', 'tinyolap.storage'],
    project_urls={
        'Homepage': 'https://tinyolap.com',
        'Documentation': 'https://tinyolap.com/docs',
        'GitHub': 'https://github.com/Zeutschler/tinyolap',
    },
)


# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import inspect
import io
import json
import random
import string
import tokenize
import types

from tinyolap.exceptions import RuleException
from tinyolap.rules import RuleScope, RuleInjectionStrategy


class ModuleCode:
    def __init__(self, module, name: str, code: str, is_custom_module: bool = True):
        self._name = name
        self._is_custom_module: bool = is_custom_module
        self._code: str = code
        self._module = module

    def __hash__(self):
        return hash(repr(self))

    def __repr__(self):
        return self._code

    def __eq__(self, other):
        if not isinstance(other, ModuleCode):
            return False
        return self._code == other.code

    def contains(self, fragment: str) -> bool:
        if self._code.find(fragment) > -1:
            return True

    @property
    def module(self):
        return self._module

    @module.setter
    def module(self, value):
        self._module = value

    @property
    def name(self):
        return self._name

    @property
    def code(self):
        return self._code

    @property
    def is_custom_module(self):
        return self._is_custom_module


class FunctionCode:
    """
    Represents a code fragment for a single (Rule) function.
    """

    def __init__(self, function, name: str, module, code: str, cube: str = None, trigger=None,
                 scope: RuleScope = None, injection: RuleInjectionStrategy = None):
        self._name = name
        self._module: str = module
        self._code: str = code
        self._cube = cube
        self._trigger = trigger
        self._scope: RuleScope = scope
        self._injection: RuleInjectionStrategy = injection
        self._function = function

    def __hash__(self):
        return hash(repr(self))

    def __repr__(self):
        return self._code

    def __eq__(self, other):
        if not isinstance(other, FunctionCode):
            return False
        return self._code == other.code

    @property
    def function(self):
        """The (Python) function object."""
        return self._function

    @function.setter
    def function(self, value):
        """The (Python) function object."""
        self._function = value

    @property
    def name(self) -> str:
        """The name of the function."""
        return self._name

    @property
    def code(self) -> str:
        """The scource code of the function."""
        return self._code

    @property
    def module(self) -> str:
        """
        A ModuleCode object that represents the code module in which the function was defined.
        Only available if the injection strategy is defined for module scope or higher.
        """
        return self._module

    def has_module(self) -> bool:
        """Identifies if the FunctionCode object has an assigned ModuleCode object."""
        return bool(self._module)

    @property
    def cube(self) -> str:
        """The cube the function has been assigned to."""
        return self._cube

    @property
    def trigger(self):
        """The trigger the function"""
        return self._trigger

    @property
    def scope(self) -> RuleScope:
        """The rule scope of the function."""
        return self._scope

    @property
    def injection(self) -> RuleInjectionStrategy:
        """The injection strategy defined for the function.
        If no injection strategy is defined, then the rule function
        will not be initialized when a database will be reloaded.
        """
        return self._injection

    def signature(self) -> str:
        """Returns the signature of the FunctionCode object.
        :return : The signature (a json string) of the FunctionCode object.
        """
        return json.dumps({"cube": self._cube, "trigger": str(self._trigger)})

    @property
    def is_valid(self) -> bool:
        """
        Identifies if the function represents a valid rule. This is the case if the rule takes
        at least one argument and *cube* and a *trigger* are defined. Invalid rule functions will
        not be initialized when a database will be reloaded.
        :return:
        """
        if self._function:
            args = str(inspect.signature(self._function))
            args = args[1: len(args) - 1].split(",")
            return bool(self._cube) and bool(self._trigger) and (len(args) > 0)
        return False


class CodeManager:
    """
    Manages code fragments (from functions or modules) used for TinyOlap rules.
    Mainly serialization, deserialization and loading (instantiation) of code.
    """

    def __init__(self):
        self.functions: dict[str, FunctionCode] = {}
        self.modules: dict[str, ModuleCode] = {}
        self._save_pending: bool = False

    @property
    def pending_changes(self) -> bool:
        """Identifies if the code manager conatins pending changes. Used for database persistence."""
        return self._save_pending

    @pending_changes.setter
    def pending_changes(self, value: bool):
        """Identifies if the code manager conatins pending changes. Used for database persistence."""
        self._save_pending = value

    def clear(self):
        """
        Clears the code manager and all contained code fragments.

        .. note::
            Note: Call this method before you want update your rules, to prohibite "ghost rules".
            The signature of each rule is defined by it's *cube* and the *trigger* assignment,
            not by the name or contents of the rule. Even slight changes might to the signature
            will create a new code fragment, then old code fragments are still active and may
            be handled instead of the new code.
        :return:
        """
        self.functions = {}
        self.modules = {}
        self._save_pending = True

    def get_functions(self, cube: str, scope: RuleScope = None) -> list[FunctionCode]:
        """
        Returns a list of FunctionCode objects matching the given parameters.
        :param cube: The cube name to be filtered.
        :param scope: (optional) the rulescope to be filtered.
        :return: a list of FunctionCode object matching the cube name and rule scope.
        """
        if scope:
            return [f for f in self.functions.values() if f.cube == cube and f.scope == scope]
        else:
            return [f for f in self.functions.values() if f.cube == cube]

    def register_function(self, function, cube: str = None, trigger: list[str] = None,
                          scope: RuleScope = None, injection: RuleInjectionStrategy = None,
                          code: str = None) -> FunctionCode:
        """
        Registers a Python function.
        :param code: (optional) the scoure code of the function
        :param function: The function to be registed.
        :param cube: (optional) the cube the function is assigned to.
        :param trigger: (optional) the trigger the function is assigned to.
        :param scope: (optional) scope of the function.
        :param injection: (optional) injectionstrategy of the function,
        :return: A FunctionCode object representing the function.
        """
        is_valid = True

        is_lambda = repr(function).find("<lambda>") != -1
        if not inspect.isroutine(function):
            raise TypeError(f"Argument 'function' is not a Python function, type id '{type(function)}'.")

        # validate settings from @rule decorator (if available)
        # todo: refactoring required > the following code should be moved to the FunctionCode class.
        if hasattr(function, "cube"):
            cube = function.cube
        if hasattr(function, "trigger"):
            trigger = function.trigger
            if trigger:
                if type(trigger) is str:
                    trigger = [trigger, ]
                elif type(trigger) is tuple:
                    trigger = [trigger, ]
                if not type(trigger) is list:
                    trigger = []
            else:
                trigger = []
        else:
            if type(trigger) is str:  # maybe a lazy user forgot to put the trigger into squared brackets
                trigger = [trigger, ]
            elif type(trigger) is tuple:
                trigger = [trigger, ]
            if not type(trigger) is list:
                trigger = []

        if hasattr(function, "scope"):
            scope = function.scope
        if hasattr(function, "injection"):
            injection = function.injection
        elif not injection:
            injection = RuleInjectionStrategy.NO_INJECTION

        # setup the function code object
        if is_lambda:
            name = self._generate_unique_name("tinyolap_lambda_rule_")
            code = self._lambda_to_function_code(function, name, cube, trigger, scope, injection)
        else:
            name = str(function).split(" ")[1]
            if not code:
                try:
                    code = inspect.getsource(function)
                except Exception as err:
                    code = None
        module = inspect.getmodule(function)
        if module:
            module_name = module.__name__
        else:
            module_name = None
        fc = FunctionCode(function, name=name, module=module_name, code=code, cube=cube,
                          trigger=trigger, scope=scope, injection=injection)

        if injection > RuleInjectionStrategy.FUNCTION_INJECTION:
            # We need to remember (persists) the entire module, so...
            module_code = inspect.getsource(module)
            # setup and register (or override) the module code object
            mc = ModuleCode(module=module, name=module_name, code=module_code)
            self.modules[module_name] = mc

        # register (or override) the function
        self.functions[fc.signature()] = fc

        self._save_pending = True
        return fc

    def _lambda_to_function_code(self, function, name: str, cube: str, trigger: list[str],
                                 scope: RuleScope, injection: RuleInjectionStrategy) -> str:
        """Converts a lambda expression into a function body."""

        # get source and extract the 'lambda' expression by tokenization
        source = " ".join([line.strip().removesuffix("\\").strip()
                           for line in inspect.getsource(function).splitlines()])
        tokens = [token for token in tokenize.generate_tokens(io.StringIO(source).readline)]
        lambda_token = None
        end_of_signature_token = None
        terminal_token = None
        for i in range(len(tokens)):
            if tokens[i].string == "lambda" and tokens[i].type == 1:  # type 1 := NAME
                lambda_token = tokens[i]
                # from here we need beyond the next ':' token
                for j in range(i + 1, len(tokens)):
                    if tokens[j].string == ":" and tokens[j].type == 54:  # type 54 := OP
                        end_of_signature_token = tokens[j]
                        # from here we need go up the 'cube' NAME token followed by the '=' OP token. If defined...
                        for k in range(j + 1, len(tokens)):
                            if tokens[k].string == "cube" and tokens[k].type == 1 and \
                                    tokens[k + 1].string == "=" and tokens[k + 1].type == 54:
                                terminal_token = tokens[k - 1]
                                break

                        # ...not found? Search for other terminals.
                        if not terminal_token:
                            end_of_lambda = self._get_end_of_lambda(source, end_of_signature_token.end[1] + 1)
                            if end_of_lambda > 0:
                                for k in range(j + 1, len(tokens)):
                                    if tokens[k].start[1] > end_of_lambda:
                                        terminal_token = tokens[k - 1]
                                        break
                                if not terminal_token:
                                    terminal_token = tokens[-1]
                        break
                break

        if not lambda_token or not end_of_signature_token or not terminal_token:
            raise RuleException("Unable to extract rule from 'lambda' expression.")

        arguments = source[lambda_token.end[1] + 1: end_of_signature_token.start[1]].strip()
        lambda_source = source[end_of_signature_token.end[1] + 1: terminal_token.start[1]].strip()
        source = lambda_source

        indentation = " " * 4
        code = f'@rule(cube="{cube}", trigger={str(trigger)}, scope={str(scope)}, injection={str(injection)})\n'
        code += f'def {name}({arguments}):\n'
        code += f'{indentation}return {source}\n'
        return code

    @staticmethod
    def _get_end_of_lambda(source: str, start: int) -> int:
        # rough evaluation the end of lambda statement.
        depth = 0
        inquote = False
        indquote = False
        for pos, char in enumerate(source[start:]):
            if char in "([{":
                depth += 1
                continue
            if char in ")]}":
                depth -= 1
                continue
            if depth > 0:
                continue
            if not indquote and char == "'":
                inquote = not inquote
            if not inquote and char == '"':
                indquote = not indquote

            if not inquote and not indquote and char == ",":
                return start + pos

        if depth == 0:
            return len(source) - 1

    def build(self):
        """
        Builds (loads and instantiates) all functions available in the code manager.
        :return:
        """
        # todo: Take care about unloading (reloading) module to not
        #      see: https://stackoverflow.com/questions/437589/how-do-i-unload-reload-a-python-module

        # 1. collect all 'modules' that need to be instantiated (no doublets please)
        modules_code = {}
        # if any(self.modules):
        # some_module = list(self.modules.values())[0].module
        for key, function in [(key, value) for key, value in self.functions.items()
                              if value.injection >= RuleInjectionStrategy.MODULE_INJECTION]:
            modules_code[self.modules[function.module].code] = self.modules[function.module]
        # instantiate the code for all distinct modules
        for module_code in modules_code.values():
            name = module_code.name
            source = module_code.code
            self.modules[module_code.name].module = self._code_to_module(name, source)
        # get reference of 'functions' from newly instatiated modules and assign them to the function code objects.
        for key, function in [(key, value) for key, value in self.functions.items()
                              if value.injection >= RuleInjectionStrategy.MODULE_INJECTION]:
            module = self.modules[function.module].module
            function.function = getattr(module, function.name)

        # 2. collect all 'isolated function' that need to be add to a default menu.
        #    ...those that have injection set to RuleInjectionStrategy.FUNCTION_INJECTION
        isolated_function_keys = [key for key, value in self.functions.items()
                                  if value.injection == RuleInjectionStrategy.FUNCTION_INJECTION]
        if isolated_function_keys:
            # generate 1 Python module containing all isolated rule functions
            module_source = self._default_rule_module_source()
            for key in isolated_function_keys:
                module_source += "\n\n" + self.functions[key].code
            module_source += "\n"

            # instantiate code and assign functions
            module_name = self._generate_unique_name("tinyolap_rules_module_")
            module = self._code_to_module(name=module_name, code=module_source)
            for key in isolated_function_keys:
                self.functions[key]._module = module_name
                function_name = self.functions[key].name
                self.functions[key]._function = getattr(module, function_name)

    def to_json(self) -> str:
        """
        Returns a string in json format containing all registered code fragments.
        :return: Json representing the code fragements contained in the code manager.
        """
        code = {}
        modules = []
        for module in self.modules.values():
            mod = {"name": module.name,
                   "is_custom": module.is_custom_module,
                   "code": module.code}
            modules.append(mod)
        code["modules"] = modules
        functions = []
        for function in self.functions.values():
            mod = {"name": function.name,
                   "function": None,
                   "module": function.module,
                   "cube": function.cube,
                   "trigger": function.trigger,
                   "scope": function.scope,
                   "injection": function.injection,
                   "code": function.code}
            functions.append(mod)
        code["functions"] = functions

        return json.dumps(code, indent=4)

    def from_json(self, data: str, build_code: bool = True) -> CodeManager:
        """
        Inititializes the code manager from code fragements supplied in json format.
        :param data: The json data to read from.
        :param build_code: Flag that defines is the code should be directly build and instantiated.
        """
        self.modules = {}
        self.functions = {}
        if not data:
            return self

        config = json.loads(data)
        for function in config["functions"]:
            self.functions[function["name"]] = FunctionCode(
                function=None,
                name=function["name"],
                module=function["module"],
                cube=function["cube"],
                trigger=function["function"],
                scope=function["scope"],
                injection=function["injection"],
                code=function["code"])
        for module in config["modules"]:
            self.modules[module["name"]] = ModuleCode(
                module=None,
                name=module["name"],
                code=module["code"],
                is_custom_module=module["is_custom"])

        if build_code:
            self.build()

        return self

    @staticmethod
    def _generate_unique_name(prefix: str = "tinolap_", length: int = 8):
        """Generate a somehow unique module name, like this 'tinyolap_gwef65lg'"""
        return prefix + ''.join(random.choices(string.ascii_lowercase + string.digits, k=length))

    @staticmethod
    def _code_to_module(name: str, code: str):
        """
        Create a new Python code module instance.
        :param name: name of the code module to instantiate.
        :param code: Source code of the code module (UTF-8).
        :return: A python code modul.
        """
        module = types.ModuleType(name)
        exec(code, module.__dict__)
        return module

    @staticmethod
    def _default_rule_module_source() -> str:
        """
        Returns a default module body (mainly imports) for rules
        that are injected on method level (FUNCTION_INJECTION = 1).
        """
        return '\n'.join([
            "import math, cmath, statistics, decimal, fractions, random, datetime, time, re, json",
            "from tinyolap.decorators import *",
            "from tinyolap.database import *",
            "from tinyolap.dimension import *",
            "from tinyolap.cube import *",
            "from tinyolap.rules import *",
        ])



# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os

from tinyolap.case_insensitive_dict import CaseInsensitiveDict
from tinyolap.database import Database


class Server:
    """
    Represents a TinyOlap server instance serving one or more databases.
    Only needed when multiple databases need to be served through a single object.
    """

    class Settings:
        version = "0.1"

    UNDEFINED = "undefined"
    DB_FILE_EXT = ".db"
    DB_DEFAULT_FOLDER = "db"

    def __init__(self):
        self._databases: CaseInsensitiveDict[str, Database] = CaseInsensitiveDict()
        self.__initialize()

    def reinitialize(self):
        """Re-initializes the server. This forces all allocated resources to be released."""
        self.__initialize()

    def __initialize(self):
        """Initializes the server for first use."""
        self._databases = CaseInsensitiveDict()
        files = self.get_existing_database_files()
        for file in files:
            database = Database(file)
            self._databases[database.name] = database

    # region database access via indexing/slicing
    def __getitem__(self, args):
        if type(args) is str:
            dbs = [db.name for db in self._databases.values()]
            if args in self._databases:
                return self._databases[args]
            raise KeyError(f"A database named '{args}' is not registered on the server.")
        raise KeyError(f"Invalid database name '{str(args)}'.")

    def __delitem__(self, args):
        self.deletes_database(args[0])
    # endregion

    def open_database(self, database_file: str):
        """
        Opens an existing database.

        Parameters
        ----------
        database_file : str
            A full qualified path to database file.
        """
        database = Database(database_file)
        if not database.open(database_file):
            return False
        if database.name in self._databases:
            raise KeyError(f"Method 'open_database()' failed. "
                                    f"A database named '{database.name}' already exists.")
        self._databases[database.name] = database
        return True

    def create_database(self, name: str, in_memory: bool = True, overwrite_existing: bool = False):
        """
        Creates a new database. If parameter ``in_memory``is set to ``True``,
        no database file will be created. For that case, all changes made to the database
        and data entered or imported will be lost after your application has shut down.

        If parameter ``in_memory``is set to ``False`` (the default settings), then a
        database file will be created in the folder 'db' aside your script root.
        If a database with the same name already exists, then the old database file will
        be renamed and a new database file will be created. Renaming simply appends a
        timestamp.

        .. danger::
           If parameter ``in_memory``is set to ``True`` and parameter ``overwrite_existing``
           is also set to ``True`` then an any existing database file will be deleted/overwritten.

        :param overwrite_existing: Identifies that an already existing database file with the
        same name will be overwritten.
        :param name: The name of the database to be created.
        Special characters are not supported for database names.
        :param in_memory: Defines if the database should operate in-memory only,
        without persistence (all data will be lost after your application will shut down)
        or, if set to ``False``, with an SQLite file storage_provider.
        """
        if not name in self._databases:
            database = Database(name, in_memory)
            self._databases[database.name] = database
            return database
        else:
            raise KeyError(f"A database named '{name}' is already registed in the server.")

    def add_database(self, database: Database) -> Database:
        """
        Adds a database to the server
        :param database: The database to be added.
        """
        if not database.name in self._databases:
            self._databases[database.name] = database
            return database
        else:
            raise KeyError(f"A database named '{database.name}' is already registed in the server.")

    def deletes_database(self, name: str):
        """Deletes a database from the filesystem. Handle with care."""
        return NotImplemented

    @staticmethod
    def get_database_folder():
        return os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            Server.DB_DEFAULT_FOLDER)

    @staticmethod
    def get_database_filename(name: str):
        return os.path.join(Server.get_database_folder(), name + Server.DB_FILE_EXT)

    @staticmethod
    def get_existing_database_files():
        files = []
        database_directory = Server.get_database_folder()
        if not os.path.exists(database_directory):
            os.mkdir(database_directory)
            return files
        for file in os.listdir(database_directory):
            if file.lower().endswith(Server.DB_FILE_EXT):
                files.append(os.path.join(database_directory, file))
        return files

    # region functions
    def _register(self, func, database: str, cube: str, pattern: list[str]):
        if database not in self._databases:
            raise KeyError(f"Database '{database}' not found.")
        if cube not in self._databases[database].cubes:
            raise KeyError(f"Cube '{cube}' of database '{database}' not found.")
        self._databases[database].cubes[cube]._register(func, pattern)
    # endregion



# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations
import datetime
import bisect
from enum import IntEnum

from tinyolap.commands import *


class HistoryMode(IntEnum):
    OFF = 0
    SESSION = 1
    PERSIST = 2


# class TimeMachineCommandEnum(Enum):
#     UNDEFINED = 0
#     SET = 1
#     ADD_DIM = 10
#     DEL_DIM = 11
#     ALTER_DIM = 12
#     ADD_CUBE = 50
#     DEL_CUBE = 51


class HistoryPartition:
    """Represents a partition of a History log.
    The HistoryLog is segmented into partitions to support log persistence over multiple usage sessions
    of a database, e.g. over multiple days, weeks, months or years. Only the last relevant partitions should
    be loaded into memory (by default)."""

    def __init__(self):
        self._log: list[Command] = []
        self._log_ts: list[datetime] = []
        self._first: datetime = datetime.datetime.now()
        self._last: datetime = self._first
        self._cursor: int = 0

    @property
    def first(self) -> datetime:
        return self._first

    @property
    def last(self) -> datetime:
        return self._last

    @property
    def cursor(self) -> int:
        """Returns the actual cursor of the partition. If the History is not is use,
        the cursor points to the end of the partition, otherwise to the entry next to be
        undone. If all entries of the partition are already undone, then cursor will
        return -1."""
        return self._cursor

    def reset(self):
        """Resets the cursor of the partition."""
        self._cursor = len(self._log) - 1

    def get_and_move_previous(self):
        """
        Returns the log entry at the current cursor position and moves the cursor to the previous log entry.
        :return: A tuple containing 1st a boolean indicating if the cursor is already at the first position,
                 2nd the timestamp of the log entry, 3rd the data of the log entry.
        """
        if self._cursor < 0:
            raise IndexError("Cursor is already at the begin.")
        self._cursor -= 1
        return self._cursor > 0, self._log_ts[self._cursor + 1], self._log[self._cursor + 1]

    def get_and_move_next(self):
        """
        Returns the log entry at the current cursor position and moves the cursor to the next log entry.
        :return: A tuple containing 1st a boolean indicating if the cursor is already at the end of the log,
                 2nd the timestamp of the log entry, 3rd the data of the log entry.
        """
        if self._cursor >= len(self._log):
            raise IndexError("Cursor is already at the end.")
        self._cursor += 1
        return self._cursor >= len(self._log), self._log_ts[self._cursor - 1], self._log[self._cursor - 1]

    def __len__(self):
        return len(self._log)

    def any(self, from_timestamp: datetime = datetime.datetime.min,
            to_timestamp: datetime = datetime.datetime.max):
        """Evaluates if any log entries exists that falls into a given time frame.

        :param from_timestamp: Form timestamp. The earliest timestamp to consider.
        :param to_timestamp: To timestamp. The latest timestamp to consider.
        :return: ``True``, if at least one log entry exists, ``False`` otherwise.
        """
        return self.count(from_timestamp, to_timestamp) > 0

    def count(self, from_timestamp: datetime = datetime.datetime.min,
              to_timestamp: datetime = datetime.datetime.max):
        """Evaluates the number of log entries that fall into a given time frame.

        :param from_timestamp: Form timestamp. The earliest timestamp to consider.
        :param to_timestamp: To timestamp. The latest timestamp to consider.
        :return: The number of log entries within the given timeframe.
        """
        if to_timestamp < self._first or from_timestamp > self._last:
            return 0
        if from_timestamp <= self._first and to_timestamp >= self._last:
            return len(self._log)

        first = bisect.bisect_left(self._log_ts, from_timestamp)
        last = bisect.bisect_right(self._log_ts, to_timestamp, lo=first)
        return last - first

    def append(self, command: Command):
        """
        Appends data to the log partition.
        :param command: The command to be added to the log.
        """
        if len(self._log) == 0:
            self._first = command.timestamp
            self._last = self._first
        else:
            self._last = command.timestamp
        self._log_ts.append(self._last)
        self._log.append(command)

    def clear(self, from_timestamp: datetime = datetime.datetime.min,
              to_timestamp: datetime = datetime.datetime.max):
        """
        Clears the log. If none of the optional parameters ``from_timestamp`` and ``to_timestamp`` will defined,
        then the entire log will be cleared. Otherwise only the log entries of the given timeframe will be deleted.
        :param from_timestamp: Form timestamp. The earliest timestamp to consider.
        :param to_timestamp: To timestamp. The latest timestamp to consider.
        """
        if to_timestamp < self._first or from_timestamp > self._last:
            return  # nothing to delete

        if from_timestamp <= self._first and to_timestamp >= self._last:
            # delete all
            self._log = []
            self._log_ts = []
            self._first: datetime = datetime.datetime.now()
            self._last: datetime = self._first
            return

        # delete specific entries from time frame.
        first = bisect.bisect_left(self._log_ts, from_timestamp)
        last = bisect.bisect_right(self._log_ts, to_timestamp, lo=first)
        self._log = self._log[:first] + self._log[last:]
        self._log_ts = self._log_ts[:first] + self._log_ts[last:]

    def pop(self):
        """Removes and returns the last entry from the partition."""
        raise NotImplementedError()

    def peek(self):
        """Returns the last entry from the partition without removing it."""
        raise NotImplementedError()

    def __getitem__(self, item) -> (datetime, Command):
        if item is int:
            return self._log_ts[item], self._log[item]
        raise KeyError("Int index expected.")


class HistoryCursor:
    def __init__(self, log: HistoryLog):
        self._log = log
        self._partition: int = len(log._partitions) - 1
        self._index: int = len(log._partitions[-1]) - 1
        self._first = datetime.datetime.min
        self._last = self._first

    def __len__(self):
        return len(self._log)

    @property
    def length(self) -> int:
        """Returns the length of the history log."""
        return len(self._log)

    @property
    def earliest_timestamp(self) -> datetime:
        """
        Returns the first timestamp available in the history.
        If the history is empty, the current date-time will be returned.
        """
        if len(self._log._partitions[0]):
            timestamp = self._log._partitions[0]._log_ts[0]
        else:
            timestamp = datetime.datetime.now()
        return timestamp

    @property
    def latest_timestamp(self) -> datetime:
        """
        Returns the latest timestamp available in the history.
        If the history is empty, the current date-time will be returned.
        """
        last_partition = len(self._log._partitions) - 1
        if last_partition >= 0:
            len_last_partition = len(self._log._partitions[last_partition]._log_ts)
            if len_last_partition == 0:
                if last_partition == 0:
                    timestamp = datetime.datetime.now()
                else:
                    partition = last_partition - 1
                    timestamp = self._log._partitions[last_partition]._log_ts[-1]
            else:
                timestamp = self._log._partitions[last_partition]._log_ts[-1]
        else:
            timestamp = datetime.datetime.now()
        return timestamp


class HistoryLog:
    """A log to support the time machine capabilities for TinyOlap databases."""

    def __init__(self):
        self._partition_max_size: int = 100_000  # the max. size of a partition, in number of records
        self._partitions: list[HistoryPartition] = [HistoryPartition()]
        self._current_partition: int = 0
        self._cursor_partition: int = 0
        self._cursor = HistoryCursor(self)

    @property
    def first(self) -> datetime:
        """Returns the first timestamp contained in the HistoryLog."""
        return self._partitions[0].first

    @property
    def last(self) -> datetime:
        """Returns the last timestamp contained in the HistoryLog."""
        return self._partitions[self._current_partition].last

    def __len__(self):
        return sum([len(partition) for partition in self._partitions])

    def clear(self):
        """Clears the time machine log."""
        self._partitions: list[HistoryPartition] = [HistoryPartition()]
        self._current_partition: int = 0

    def is_empty(self) -> bool:
        return len(self) == 0

    def append(self, data):
        self._partitions[self._current_partition].append(data)

    def get_and_move_previous(self):
        """
        Returns the log entry at the current cursor position and moves the cursor to the previous log entry.
        :return: A tuple containing 1st a boolean indicating if the cursor is already at the first position,
                 2nd the timestamp of the log entry, 3rd the data of the log entry.
        """
        if self._cursor_partition < 0:
            raise IndexError("Cursor is already at the begin.")
        has_previous, timestamp, data = self._partitions[self._cursor_partition].get_and_move_previous()
        if not has_previous:
            self._cursor_partition -= 1
        return self._cursor_partition >= 0, timestamp, data

    def get_and_move_next(self):
        """
        Returns the log entry at the current cursor position and moves the cursor to the next log entry.
        :return: A tuple containing 1st a boolean indicating if the cursor is already at the end of the log,
                 2nd the timestamp of the log entry, 3rd the data of the log entry.
        """
        if self._cursor_partition > len(self._partitions):
            raise IndexError("Cursor is already at the end.")
        has_next, timestamp, data = self._partitions[self._cursor_partition].get_and_move_next()
        if not has_next:
            self._cursor_partition += 1
        return self._cursor_partition <= len(self._partitions), timestamp, data

    def reset(self):
        """Resets the cursor for the log."""
        for partition in self._partitions:
            partition.reset()
        self._cursor_partition = len(self._partitions) - 1

    def entries(self, reverse: bool = True):
        """Return the entries of the log, implemented as an iterator. Calling this method will reset the cursor.

        :param reverse: If set to ``True``(the default value), then the log will be iterated in reverse order,
                        meaning latest entries first.
        :return: The entries from the log.
        """
        self.reset()
        while self._cursor_partition >= 0:
            has_previous, timestamp, data = self._partitions[self._cursor_partition].get_and_move_previous()
            if not has_previous:
                self._cursor_partition -= 1
            yield self._cursor_partition >= 0, timestamp, data


class History:
    def __init__(self, database):
        self._database = database
        self._log = HistoryLog()
        self._cursor = HistoryCursor(self._log)
        self._mode = HistoryMode.OFF

    @property
    def mode(self) -> HistoryMode:
        """Returns the time machine mode."""
        return self._mode

    @mode.setter
    def mode(self, value: HistoryMode):
        """Sets the time machine mode."""
        self._mode = value

    @property
    def is_enabled(self):
        return not self._mode == HistoryMode.OFF

    def append(self, command: Command):
        """
        Appends a command to the history.
        :param command: The command to append.
        """
        if self._mode is HistoryMode.OFF:
            return

        self._log.append(command)

    def append_cube_set(self, cube: str, address, value_before, value_after):
        """
        Appends a cube cell value change to the the History log.
        :param address: address of the cell to be added
        :param value_before: The previous value of the cell.
        :param value_after: The new value of the cell.
        """
        if self._mode is HistoryMode.OFF:
            return

        idx_address = self._database.cubes[cube]._address_to_idx_address(address)
        cmd = CubeSetCommand(cube, idx_address, value_before, value_after)
        self._log.append(cmd)

    def undo_to(self, timestamp: datetime.datetime = datetime.datetime.min):
        """
        Undo all executed commands up to a certain point in time. If the timestamp
        is higher then the last timestamp available in the history log, then no undo will be executed.
        If the timestamp is earlier than the first timestamp available in the history log,
        then the entire history log will be undone.

        :param timestamp: The point in time up to when the executed commands should be undone.
        :return: True if the undo was processed successfully. False otherwise.
        """
        pass

    def redo_to(self, timestamp: datetime.datetime = datetime.datetime.max):
        """
        Redo all undone executed commands up to a certain point in time. If the timestamp
        is earlier then the current cursor position within the history log (state of undo's),
        then no redo will be executed.
        If the timestamp is higher than the last timestamp available in the history log,
        then all undo's will be redone.

        :param timestamp: The point in time up to when the executed commands should be redone.
        :return: True if the undo was processed successfully. False otherwise.
        """
        pass

    def undo(self, count: int = 1):
        """
        Undo a certain number of executed commands.
        If parameter ``count`` is set to -1 or to a number higher than the number of available
        commands in the history log, then all commands will be undone.

        :param count: Number of commands to be undone, default value is 1 undo. If parameter ``count`` is set to -1 or
        to a number higher than the number of available commands in the history log, then all commands will be undone.
        :return: ``True`` if the undo was processed successfully. ``False`` otherwise.
        """
        for has_prev, timestamp, command in self._log.entries():
            self._process(command)
            count -= 1
            if count == 0:
                break

    def redo(self, count: int = 1):
        """
        Redo a certain number of executed commands.
        If parameter ``count`` is set to -1 or to a number higher than the number of available
        commands in the history log, then all commands will be undone.

        :param count: Number of commands to be undone, default value is 1 redo. If parameter ``count`` is set to -1 or
        to a number higher than the number of available commands in the history log, then all commands will be undone.
        :return: ``True`` if the redo was processed successfully. ``False`` otherwise.
        """
        pass

    def cursor_position(self) -> int:
        """
        Returns the current position of the history log cursor. If the history is at the current state,
        meaning no commands have been undone, then 0 will be returned. Otherwise the number
        currently undone commands will be returned.
        """
        pass

    def cursor_timestamp(self) -> int:
        """
        Returns the current timestamp at the history log cursor. This is the timestamp of the
        next to be undone command. If all commands have already been undone,
        then ``datetime.datetime.min`` will be returned.
        """
        pass

    def _process(self, command: Command, undo: bool = True):
        if undo:
            command.undo(self._database)
        else:
            command.redo(self._database)
